# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client for handling a data storage."""
import functools
import os
import re
import shlex
import tempfile
from collections import defaultdict
from pathlib import Path
from shutil import move, which
from subprocess import PIPE, STDOUT, call, check_output, run

import attr
import pathspec
from werkzeug.utils import cached_property

from renku.core import errors
from renku.core.utils.file_size import parse_file_size

from .git import _expand_directories
from .repository import RepositoryApiMixin

# Batch size for when renku is expanding a large list
# of files into an argument string.
ARGUMENT_BATCH_SIZE = 100


def check_external_storage_wrapper(fn):
    """Check availability of external storage on methods that need it.

    :raises: ``errors.ExternalStorageNotInstalled``
    :raises: ``errors.ExternalStorageDisabled``
    """
    # noqa
    @functools.wraps(fn)
    def wrapper(self, *args, **kwargs):
        if not self.check_external_storage():
            pass
        else:
            return fn(self, *args, **kwargs)

    return wrapper


@attr.s
class StorageApiMixin(RepositoryApiMixin):
    """Client for handling a data storage."""

    external_storage_requested = attr.ib(default=True)
    """External storage (e.g. LFS) requested for Renku command."""

    RENKU_LFS_IGNORE_PATH = '.renkulfsignore'
    """.gitignore like file specifying paths that are not tracked in LFS."""

    _CMD_STORAGE_INSTALL = ['git', 'lfs', 'install', '--local']

    _CMD_STORAGE_TRACK = ['git', 'lfs', 'track', '--']

    _CMD_STORAGE_UNTRACK = ['git', 'lfs', 'untrack', '--']

    _CMD_STORAGE_CLEAN = ['git', 'lfs', 'clean']

    _CMD_STORAGE_CHECKOUT = ['git', 'lfs', 'checkout']

    _CMD_STORAGE_PULL = ['git', 'lfs', 'pull', '-I']

    _CMD_STORAGE_MIGRATE_INFO = [
        'git', 'lfs', 'migrate', 'info', '--top', '42000'
    ]

    _CMD_STORAGE_LIST = ['git', 'lfs', 'ls-files', '-n']

    _CMD_STORAGE_STATUS = ['git', 'lfs', 'status']

    _LFS_HEADER = 'version https://git-lfs.github.com/spec/'

    @cached_property
    def storage_installed(self):
        """Verify that git-lfs is installed and on system PATH."""
        return bool(which('git-lfs'))

    @cached_property
    def storage_installed_locally(self):
        """Verify that git-lfs is installed for the project."""
        repo_config = self.repo.config_reader(config_level='repository')
        return repo_config.has_section('filter "lfs"')

    def check_external_storage(self):
        """Check if repository has external storage enabled.

        :raises: ``errors.ExternalStorageNotInstalled``
        :raises: ``errors.ExternalStorageDisabled``
        """
        storage_installed = (
            self.storage_installed_locally and self.storage_installed
        )
        if self.external_storage_requested and not storage_installed:
            raise errors.ExternalStorageDisabled(self.repo)

        if self.storage_installed_locally and not self.storage_installed:
            raise errors.ExternalStorageNotInstalled(self.repo)

        return storage_installed

    @cached_property
    def renku_lfs_ignore(self):
        """Gets pathspec for files to not add to LFS."""
        ignore_path = self.path / self.RENKU_LFS_IGNORE_PATH
        if not os.path.exists(ignore_path):
            return pathspec.PathSpec.from_lines('renku_gitwildmatch', [])
        with ignore_path.open('r') as f:
            return pathspec.PathSpec.from_lines('renku_gitwildmatch', f)

    @property
    def minimum_lfs_file_size(self):
        """The minimum size of a file in bytes to be added to lfs."""
        size = self.get_value('renku', 'lfs_threshold') or '100kb'

        return parse_file_size(size)

    def init_external_storage(self, force=False):
        """Initialize the external storage for data."""
        try:
            call(
                self._CMD_STORAGE_INSTALL + (['--force'] if force else []),
                stdout=PIPE,
                stderr=STDOUT,
                cwd=self.path,
            )
        except (KeyboardInterrupt, OSError) as e:
            raise errors.ParameterError(
                'Couldn\'t run \'git lfs\':\n{0}'.format(e)
            )

    def init_repository(self, force=False, user=None):
        """Initialize a local Renku repository."""
        result = super().init_repository(force=force, user=user)

        # initialize LFS if it is requested and installed
        if self.external_storage_requested and self.storage_installed:
            self.init_external_storage(force=force)

        return result

    @check_external_storage_wrapper
    def track_paths_in_storage(self, *paths):
        """Track paths in the external storage."""
        if not self.external_storage_requested:
            return

        # Calculate which paths can be tracked in lfs
        track_paths = []
        attrs = self.find_attr(*paths)

        for path in paths:
            path = Path(path)

            # Do not track symlinks in LFS
            if path.is_symlink():
                continue

            if path.is_absolute():
                path = Path(path).relative_to(self.path)

            # Do not add files with filter=lfs in .gitattributes
            if attrs.get(str(path), {}).get('filter') == 'lfs':
                continue

            if (
                path.is_dir() and
                not any(self.renku_lfs_ignore.match_tree(str(path)))
            ):
                track_paths.append(str(path / '**'))
            elif not self.renku_lfs_ignore.match_file(str(path)):
                file_size = os.path.getsize(
                    str(os.path.relpath(self.path / path, os.getcwd()))
                )
                if file_size >= self.minimum_lfs_file_size:
                    track_paths.append(str(path))

        if track_paths:
            try:
                call(
                    self._CMD_STORAGE_TRACK + track_paths,
                    stdout=PIPE,
                    stderr=STDOUT,
                    cwd=self.path,
                )
            except (KeyboardInterrupt, OSError) as e:
                raise errors.ParameterError(
                    'Couldn\'t run \'git lfs\':\n{0}'.format(e)
                )
            return track_paths
        return []

    @check_external_storage_wrapper
    def untrack_paths_from_storage(self, *paths):
        """Untrack paths from the external storage."""
        try:
            call(
                self._CMD_STORAGE_UNTRACK + list(paths),
                stdout=PIPE,
                stderr=STDOUT,
                cwd=self.path,
            )
        except (KeyboardInterrupt, OSError) as e:
            raise errors.ParameterError(
                'Couldn\'t run \'git lfs\':\n{0}'.format(e)
            )

    @check_external_storage_wrapper
    def list_tracked_paths(self, client=None):
        """List paths tracked in lfs for a client."""
        client = client or self
        try:
            files = check_output(
                self._CMD_STORAGE_LIST, cwd=client.path, encoding='UTF-8'
            )
        except (KeyboardInterrupt, OSError) as e:
            raise errors.ParameterError(
                'Couldn\'t run \'git lfs\':\n{0}'.format(e)
            )
        files = [client.path / f for f in files.splitlines()]
        return files

    @check_external_storage_wrapper
    def list_unpushed_lfs_paths(self, client=None):
        """List paths tracked in lfs for a client."""
        client = client or self

        if (
            len(client.repo.remotes) < 1 or
            not client.repo.active_branch.tracking_branch()
        ):
            raise errors.ConfigurationError(
                'No git remote is configured for {} branch {}.'.
                format(client.path, client.repo.active_branch.name) +
                'Cleaning the storage cache would lead to a loss of data as ' +
                'it is not on a server. Please see ' +
                'https://www.atlassian.com/git/tutorials/syncing for ' +
                'information on how to sync with a remote.'
            )
        try:
            status = check_output(
                self._CMD_STORAGE_STATUS, cwd=client.path, encoding='UTF-8'
            )
        except (KeyboardInterrupt, OSError) as e:
            raise errors.ParameterError(
                'Couldn\'t run \'git lfs\':\n{0}'.format(e)
            )

        files = status.split('Objects to be committed:')[0].splitlines()[2:]
        files = [
            client.path / f.rsplit('(', 1)[0].strip()
            for f in files if f.strip()
        ]
        return files

    @check_external_storage_wrapper
    def pull_paths_from_storage(self, *paths):
        """Pull paths from LFS."""
        import math
        client_dict = defaultdict(list)

        for path in _expand_directories(paths):
            client, commit, path = self.resolve_in_submodules(
                self.repo.commit(), path
            )
            try:
                absolute_path = Path(path).resolve()
                relative_path = absolute_path.relative_to(client.path)
            except ValueError:  # An external file
                absolute_path = Path(os.path.abspath(path))
                relative_path = absolute_path.relative_to(client.path)
            client_dict[client.path].append(str(relative_path))

        for client_path, paths in client_dict.items():
            batch_size = math.ceil(len(paths) / ARGUMENT_BATCH_SIZE)
            for index in range(batch_size):
                run(
                    self._CMD_STORAGE_PULL + [
                        shlex.quote(
                            ','.join(
                                paths[index * ARGUMENT_BATCH_SIZE:(index + 1) *
                                      ARGUMENT_BATCH_SIZE]
                            )
                        )
                    ],
                    cwd=client_path,
                    stdout=PIPE,
                    stderr=STDOUT,
                )

    @check_external_storage_wrapper
    def clean_storage_cache(self, *paths):
        """Remove paths from lfs cache."""
        client_dict = defaultdict(list)
        clients = {}
        tracked_paths = defaultdict(list)
        unpushed_paths = defaultdict(list)
        untracked_paths = []
        local_only_paths = []

        for path in _expand_directories(paths):
            client, commit, path = self.resolve_in_submodules(
                self.repo.commit(), path
            )
            try:
                absolute_path = Path(path).resolve()
                relative_path = absolute_path.relative_to(client.path)
            except ValueError:  # An external file
                absolute_path = Path(os.path.abspath(path))
                relative_path = absolute_path.relative_to(client.path)

            if client.path not in tracked_paths:
                tracked_paths[client.path] = self.list_tracked_paths(client)

            if client.path not in unpushed_paths:
                u_paths = self.list_unpushed_lfs_paths(client)
                unpushed_paths[client.path] = u_paths

            if absolute_path in unpushed_paths[client.path]:
                local_only_paths.append(str(relative_path))
            elif absolute_path not in tracked_paths[client.path]:
                untracked_paths.append(str(relative_path))
            else:
                client_dict[client.path].append(str(relative_path))
                clients[client.path] = client

        for client_path, paths in client_dict.items():
            client = clients[client_path]

            for path in paths:
                with open(path, 'r') as tracked_file:
                    try:
                        header = tracked_file.read(len(self._LFS_HEADER))
                        if header == self._LFS_HEADER:
                            # file is not pulled
                            continue
                    except UnicodeDecodeError:
                        # likely a binary file, not lfs pointer file
                        pass
                with tempfile.NamedTemporaryFile(
                    mode='w+t', encoding='utf-8', delete=False
                ) as tmp, open(path, 'r+t') as input_file:
                    run(
                        self._CMD_STORAGE_CLEAN,
                        cwd=client_path,
                        stdin=input_file,
                        stdout=tmp,
                    )

                    tmp_path = tmp.name
                move(tmp_path, path)

                # get lfs sha hash
                old_pointer = client.repo.git.show('HEAD:{}'.format(path))
                old_pointer = old_pointer.splitlines()[1]
                old_pointer = old_pointer.split(' ')[1].split(':')[1]

                prefix1 = old_pointer[:2]
                prefix2 = old_pointer[2:4]

                # remove from lfs cache
                object_path = (
                    client.path / '.git' / 'lfs' / 'objects' / prefix1 /
                    prefix2 / old_pointer
                )
                object_path.unlink()

            # add paths so they don't show as modified
            client.repo.git.add(*paths)

        return untracked_paths, local_only_paths

    @check_external_storage_wrapper
    def checkout_paths_from_storage(self, *paths):
        """Checkout a paths from LFS."""
        run(
            self._CMD_STORAGE_CHECKOUT + list(paths),
            cwd=self.path,
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )

    def check_requires_tracking(self, *paths):
        """Check paths and return a list of those that must be tracked."""
        if not self.external_storage_requested:
            return

        attrs = self.find_attr(*paths)
        track_paths = []

        for path in paths:
            absolute_path = Path(os.path.abspath(self.path / path))
            path = str(path)

            # Do not track symlinks in LFS
            if absolute_path.is_symlink():
                continue

            # Do not add files with filter=lfs in .gitattributes
            if attrs.get(path, {}).get('filter') == 'lfs':
                continue

            if not absolute_path.is_dir():
                if self.renku_lfs_ignore.match_file(path):
                    continue
                if os.path.getsize(absolute_path) < self.minimum_lfs_file_size:
                    continue

                track_paths.append(path)

        return track_paths

    def check_lfs_migrate_info(self, everything=False):
        """Return list of file groups in history should be in LFS."""
        ref = ['--everything'] if everything else [
            '--include-ref', self.repo.active_branch.name
        ]

        includes = []
        excludes = []
        for p in self.renku_lfs_ignore.patterns:
            if p.regex is None:
                continue

            pattern = p.pattern.replace(os.linesep, '').replace('\n', '')
            if pattern.startswith('!'):
                pattern.replace('!', '', 1)

            if p.include:  # File ignored by LFS
                excludes.append(pattern)
            else:
                includes.append(pattern)

        if excludes:
            excludes = ['--exclude', ','.join(excludes)]
        if includes:
            includes = ['--include', ','.join(includes)]

        above = ['--above', str(self.minimum_lfs_file_size)]

        command = (
            self._CMD_STORAGE_MIGRATE_INFO + ref + above + includes + excludes
        )

        try:
            lfs_output = run(
                command,
                stdout=PIPE,
                stderr=STDOUT,
                cwd=self.path,
                universal_newlines=True
            )
        except (KeyboardInterrupt, OSError) as e:
            raise errors.GitError(
                'Couldn\'t run \'git lfs migrate info\':\n{0}'.format(e)
            )

        groups = []
        files_re = re.compile(r'(.*\s+[\d.]+\s+\S+).*')

        for line in lfs_output.stdout.split('\n'):
            match = files_re.match(line)
            if match:
                groups.append(match.groups()[0])

        return groups



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Job queues."""
from rq import Queue

from renku.service.cache.base import BaseCache

CLEANUP_QUEUE_FILES = 'cache.cleanup.files'
CLEANUP_QUEUE_PROJECTS = 'cache.cleanup.projects'

DATASETS_JOB_QUEUE = 'datasets.jobs'
MIGRATIONS_JOB_QUEUE = 'project.migrations'

QUEUES = [
    CLEANUP_QUEUE_FILES,
    CLEANUP_QUEUE_PROJECTS,
    DATASETS_JOB_QUEUE,
    MIGRATIONS_JOB_QUEUE,
]


class WorkerQueues:
    """Worker queues."""

    connection = BaseCache.cache

    @staticmethod
    def describe():
        """List possible queues."""
        return QUEUES

    @staticmethod
    def get(name):
        """Get specific queue object."""
        return Queue(name, connection=WorkerQueues.connection)



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represents a workflow template."""

import os
from bisect import bisect
from copy import copy
from functools import total_ordering
from pathlib import Path

from renku.core.models import jsonld as jsonld
from renku.core.models.cwl.types import PATH_OBJECTS
from renku.core.models.entities import Collection, CommitMixin, Entity
from renku.core.models.workflow.parameters import CommandArgument, \
    CommandInput, CommandOutput, MappedIOStream


def _entity_from_path(client, path, commit):
    """Gets the entity associated with a path."""
    client, commit, path = client.resolve_in_submodules(
        client.find_previous_commit(path, revision=commit.hexsha),
        path,
    )

    entity_cls = Entity
    if (client.path / path).is_dir():
        entity_cls = Collection

    if str(path).startswith(os.path.join(client.renku_home, client.DATASETS)):
        return client.load_dataset_from_path(path, commit=commit)
    else:
        return entity_cls(
            commit=commit,
            client=client,
            path=str(path),
        )


def _convert_cmd_binding(binding):
    """Convert a cwl argument to ``CommandArgument``."""
    return CommandArgument(position=binding.position, value=binding.valueFrom)


def _convert_cmd_input(input, client, commit):
    """Convert a cwl input to ``CommandInput``."""
    val = input.default

    if isinstance(val, list):
        val = input.inputBinding.itemSeparator.join(val)

    if input.type in PATH_OBJECTS and input.default:
        if input.inputBinding:
            prefix = input.inputBinding.prefix
            if prefix and input.inputBinding.separate:
                prefix += ' '
            return CommandInput(
                position=input.inputBinding.position,
                prefix=prefix,
                consumes=_entity_from_path(client, input.default.path, commit)
            )
        else:
            return CommandInput(
                consumes=_entity_from_path(client, input.default.path, commit),
                mapped_to=MappedIOStream(stream_type='stdin')
                if input.id == 'input_stdin' else None
            )
    else:
        prefix = input.inputBinding.prefix
        if prefix and input.inputBinding.separate:
            prefix += ' '
        return CommandArgument(
            position=input.inputBinding.position, value=val, prefix=prefix
        )


def _convert_cmd_output(output, factory, client, commit):
    """Convert a cwl output to ``CommandOutput``."""
    path = None
    mapped = None
    input_prefix = '$(inputs.'
    position = None
    prefix = None
    input_to_remove = None
    create_folder = False

    if output.outputBinding:
        if output.outputBinding.glob.startswith(input_prefix):
            input_id = output.outputBinding.glob[len(input_prefix):-1]
            inp = next(i for i in factory.inputs if i.id == input_id)
            path = inp.default
            position = inp.inputBinding.position
            prefix = inp.inputBinding.prefix
            if prefix and inp.inputBinding.separate:
                prefix += ' '
            input_to_remove = inp
        else:
            path = output.outputBinding.glob

    if output.type in MappedIOStream.STREAMS:
        path = getattr(factory, output.type)
        mapped = MappedIOStream(stream_type=output.type)

    if (((client.path / path).is_dir() and
         path in factory.existing_directories) or (
             not (client.path / path).is_dir() and
             str(Path(path).parent) in factory.existing_directories
         )):
        create_folder = True

    return CommandOutput(
        produces=_entity_from_path(client, path, commit),
        mapped_to=mapped,
        position=position,
        prefix=prefix,
        create_folder=create_folder
    ), input_to_remove


@total_ordering
@jsonld.s(
    type=[
        'renku:Run',
        'prov:Entity',
        'prov:Plan',
    ],
    context={
        'renku': 'https://swissdatasciencecenter.github.io/renku-ontology#',
        'prov': 'http://www.w3.org/ns/prov#',
    },
    cmp=False,
)
class Run(CommitMixin):
    """Represents a `renku run` execution template."""

    command = jsonld.ib(
        default=None,
        context={
            '@id': 'renku:command',
            '@type': 'http://www.w3.org/2001/XMLSchema#string',
        },
        type=str,
        kw_only=True,
    )

    process_order = jsonld.ib(
        default=None,
        context={
            '@id': 'renku:processOrder',
            '@type': 'http://www.w3.org/2001/XMLSchema#integer',
        },
        type=int,
        kw_only=True,
    )

    successcodes = jsonld.container.list(
        context='renku:successCodes', kw_only=True, type=int
    )

    subprocesses = jsonld.container.list(
        'renku.core.models.workflow.run.Run',
        context='renku:hasSubprocess',
        kw_only=True
    )

    arguments = jsonld.container.list(
        context='renku:hasArguments', kw_only=True, type=CommandArgument
    )

    inputs = jsonld.container.list(
        context='renku:hasInputs', kw_only=True, type=CommandInput
    )

    outputs = jsonld.container.list(
        context='renku:hasOutputs', kw_only=True, type=CommandOutput
    )

    @classmethod
    def from_factory(cls, factory, client, commit, path):
        """Creates a ``Run`` from a ``CommandLineToolFactory``."""
        inputs = []
        arguments = []
        outputs = [
            _convert_cmd_output(o, factory, client, commit)
            for o in factory.outputs
        ]  # TODO: handle stream!

        if outputs:
            outputs, inputs_to_remove = zip(*outputs)
            outputs = list(outputs)

            for i in inputs_to_remove:
                # remove inputs that are actually outputs
                # note: a single input can represent multiple outputs
                # in case of repetition in the cli
                if not i:
                    continue
                if i in factory.inputs:
                    factory.inputs.remove(i)

        for i in factory.inputs:
            res = _convert_cmd_input(i, client, commit)

            if isinstance(res, CommandInput):
                inputs.append(res)
            else:
                arguments.append(res)

        return cls(
            client=client,
            commit=commit,
            path=path,
            command=' '.join(factory.baseCommand),
            successcodes=factory.successCodes,
            arguments=[_convert_cmd_binding(a)
                       for a in factory.arguments] + arguments,
            inputs=inputs,
            outputs=outputs
        )

    @property
    def activity(self):
        """Return the activity object."""
        return self._activity()

    def to_argv(self):
        """Convert run into argv list."""
        argv = []

        if self.command:
            argv.extend(self.command.split(' '))

        arguments = self.inputs + self.outputs + self.arguments

        arguments = filter(lambda x: x.position, arguments)
        arguments = sorted(arguments, key=lambda x: x.position)
        argv.extend(e for a in arguments for e in a.to_argv())

        return argv

    def to_stream_repr(self):
        """Input/output stream representation."""
        stream_repr = []

        for input_ in self.inputs:
            if input_.mapped_to:
                stream_repr.append(input_.to_stream_repr())

        for output in self.outputs:
            if output.mapped_to:
                stream_repr.append(output.to_stream_repr())
        return stream_repr

    def update_id_and_label_from_commit_path(self, client, commit, path):
        """Updates the _id and _label using supplied commit and path."""
        self.client = client

        if not self.commit:
            self.commit = commit

            path = Path(os.path.abspath(path)).relative_to(self.client.path)
            self.path = path
            self._id = self.default_id()
            self._label = self.default_label()

        if len(self.subprocesses) > 0:
            for s in self.subprocesses:
                s.update_id_and_label_from_commit_path(client, commit, path)

    def add_subprocess(self, subprocess, process_order=None):
        """Adds a subprocess to this run."""
        if not process_order:
            process_order = 0
            if self.subprocesses:
                # sort subprocesses by dependencies
                process_order = bisect(self.subprocesses, subprocess)
                if process_order < len(self.subprocesses):
                    # inserted before end, recalculate orders or rest
                    for s in self.subprocesses:
                        if s.process_order >= process_order:
                            s.process_order += 1

        if any(s.process_order == process_order for s in self.subprocesses):
            raise ValueError(
                'process_order {} already exists'.format(process_order)
            )

        subprocess.process_order = process_order

        input_paths = [i.consumes.path for i in self.inputs]
        output_paths = [o.produces.path for o in self.outputs]

        for input_ in subprocess.inputs:
            if (
                input_.consumes.path not in input_paths and
                input_.consumes.path not in output_paths
            ):
                new_input = copy(input_)
                new_input.mapped_to = None

                matching_output = next((
                    o for o in self.outputs
                    if o.produces.path == new_input.consumes.path
                ), None)

                if not matching_output:
                    self.inputs.append(new_input)
                    input_paths.append(new_input.consumes.path)

        for output in subprocess.outputs:
            if output.produces.path not in output_paths:
                new_output = copy(output)
                new_output.mapped_to = None
                self.outputs.append(new_output)
                output_paths.append(new_output.produces.path)

                matching_input = next((
                    i for i in self.inputs
                    if i.consumes.path == new_output.produces.path
                ), None)
                if matching_input:
                    self.inputs.remove(matching_input)
                    input_paths.remove(matching_input.consumes.path)

        self.subprocesses.append(subprocess)

        self.subprocesses = sorted(
            self.subprocesses, key=lambda s: s.process_order
        )

    def __lt__(self, other):
        """Compares two subprocesses order based on their dependencies."""
        a_inputs = set()
        b_outputs = set()

        for i in other.inputs:
            entity = i.consumes
            for subentity in entity.entities:
                a_inputs.add(subentity.path)

        for i in self.outputs:
            entity = i.produces
            for subentity in entity.entities:
                b_outputs.add(subentity.path)

        return a_inputs & b_outputs



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convenience method to save local changes and push them to a remote server.

If you have local modification to files, you can save them using

.. code-block:: console

    $ renku save
    Username for 'https://renkulab.io': my.user
    Password for 'https://my.user@renkulab.io':
    Successfully saved:
        file1
        file2
    OK

.. warning:: The username and password for renku save are your gitlab
   user/password, not your renkulab login!

You can additionally supply a message that describes the changes that you
made by using the ``-m`` or ``--message`` parameter followed by your
message.

.. code-block:: console

    $ renku save -m "Updated file1 and 2."
    Successfully saved:
        file1
        file2
    OK

If no remote server has been configured, you can specify one by using the
``-d`` or ``--destination`` parameter. Otherwise you will get an error.

.. code-block:: console

    $ renku save
    Error: No remote has been set up for the current branch

    $ renku save -d https://renkulab.io/gitlab/my.user/my-project.git
    Successfully saved:
        file1
        file2
    OK

You can also specify which paths to save:

.. code-block:: console

    $ renku save file1
    Successfully saved:
        file1
    OK

"""

import click

from renku.core.commands.save import save_and_push


@click.command(name='save')
@click.option(
    '-m', '--message', default=None, help='The commit message to use'
)
@click.option(
    '-d',
    '--destination',
    default=None,
    help=(
        'The git remote to push to. Defaults to the remote set in git, '
        'automatically set in interactive environments'
    )
)
@click.argument('paths', type=click.Path(exists=True, dir_okay=True), nargs=-1)
@click.pass_context
def save(ctx, message, destination, paths):
    """Save and push local changes."""

    saved_paths, branch = save_and_push(
        message=message, remote=destination, paths=paths
    )

    if saved_paths:
        click.echo(
            'Successfully saved to branch {}: \n\t{}'.format(
                branch, '\n\t'.join(saved_paths)
            )
        )
    else:
        click.echo('There were no changes to save.')

    click.secho('OK', fg='green')



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service view decorators."""
import re
from functools import wraps

from flask import jsonify, request
from flask_apispec import doc
from git import GitCommandError
from marshmallow import ValidationError
from redis import RedisError
from werkzeug.exceptions import HTTPException

from renku.core.errors import MigrationRequired, RenkuException
from renku.service.cache import cache
from renku.service.config import GIT_ACCESS_DENIED_ERROR_CODE, \
    GIT_UNKNOWN_ERROR_CODE, INTERNAL_FAILURE_ERROR_CODE, \
    INVALID_HEADERS_ERROR_CODE, INVALID_PARAMS_ERROR_CODE, \
    REDIS_EXCEPTION_ERROR_CODE, RENKU_EXCEPTION_ERROR_CODE
from renku.service.serializers.headers import UserIdentityHeaders
from renku.service.views import error_response


def requires_identity(f):
    """Wrapper which indicates that route requires user identification."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kws):
        """Represents decorated function."""
        try:
            user = UserIdentityHeaders().load(request.headers)
        except (ValidationError, KeyError):
            err_message = 'user identification is incorrect or missing'
            return jsonify(
                error={
                    'code': INVALID_HEADERS_ERROR_CODE,
                    'reason': err_message
                }
            )

        return f(user, *args, **kws)

    return decorated_function


def handle_redis_except(f):
    """Wrapper which handles Redis exceptions."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except (RedisError, OSError) as e:
            error_code = REDIS_EXCEPTION_ERROR_CODE

            return jsonify(error={
                'code': error_code,
                'reason': e.messages,
            })

    return decorated_function


@handle_redis_except
def requires_cache(f):
    """Wrapper which injects cache object into view."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        return f(cache, *args, **kwargs)

    return decorated_function


def handle_schema_except(f):
    """Wrapper which handles Schema.load exceptions."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except KeyError as e:
            if (e.args and len(e.args) > 0):
                return jsonify(
                    error={
                        'code': INVALID_PARAMS_ERROR_CODE,
                        'reason': f'missing parameter "{e.args[0]}"',
                    }
                )
            raise

    return decorated_function


def handle_validation_except(f):
    """Wrapper which handles marshmallow `ValidationError`."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except ValidationError as e:
            return jsonify(
                error={
                    'code': INVALID_PARAMS_ERROR_CODE,
                    'reason': e.messages,
                }
            )

    return decorated_function


def handle_renku_except(f):
    """Wrapper which handles `RenkuException`."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except RenkuException as e:
            err_response = {
                'code': RENKU_EXCEPTION_ERROR_CODE,
                'reason': str(e),
            }

            if isinstance(e, MigrationRequired):
                err_response['migration_required'] = True

            return jsonify(error=err_response)

    return decorated_function


def handle_git_except(f):
    """Wrapper which handles `RenkuException`."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except GitCommandError as e:

            error_code = GIT_ACCESS_DENIED_ERROR_CODE \
                if 'Access denied' in e.stderr else GIT_UNKNOWN_ERROR_CODE

            # strip oauth tokens
            error_reason = format(' '.join(e.stderr.strip().split('\n')))
            error_reason_safe = re.sub(
                '^(.+)(oauth2:)(.+)(@)(.+)$', r'\1\2<token-hidden>\4\5',
                error_reason
            )

            return jsonify(
                error={
                    'code': error_code,
                    'reason': f'git error: {error_reason_safe}'
                }
            )

    return decorated_function


def accepts_json(f):
    """Wrapper which ensures only JSON payload can be in request."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        if 'Content-Type' not in request.headers:
            return jsonify(
                error={
                    'code': INVALID_HEADERS_ERROR_CODE,
                    'reason': 'invalid request headers'
                }
            )

        header_check = request.headers['Content-Type'] == 'application/json'

        if not request.is_json or not header_check:
            return jsonify(
                error={
                    'code': INVALID_HEADERS_ERROR_CODE,
                    'reason': 'invalid request payload'
                }
            )

        return f(*args, **kwargs)

    return decorated_function


def handle_base_except(f):
    """Wrapper which handles base exceptions."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except HTTPException as e:  # handle general werkzeug exception
            return error_response(e.code, e.description)

        except (Exception, BaseException, OSError, IOError) as e:
            internal_error = 'internal error'
            if hasattr(e, 'stderr'):
                internal_error += ': {0}'.format(
                    ' '.join(e.stderr.strip().split('\n'))
                )
            return error_response(INTERNAL_FAILURE_ERROR_CODE, internal_error)

    return decorated_function


def header_doc(description, tags=()):
    """Wrap additional OpenAPI header description for an endpoint."""
    return doc(
        description=description,
        params={
            'Authorization': {
                'description': (
                    'Used for users git oauth2 access. '
                    'For example: '
                    '```Authorization: Bearer asdf-qwer-zxcv```'
                ),
                'in': 'header',
                'type': 'string',
                'required': True
            },
            'Renku-User-Id': {
                'description': (
                    'Used for identification of the users. '
                    'For example: '
                    '```Renku-User-Id: sasdsa-sadsd-gsdsdh-gfdgdsd```'
                ),
                'in': 'header',
                'type': 'string',
                'required': True
            },
            'Renku-User-FullName': {
                'description': (
                    'Used for commit author signature. '
                    'For example: '
                    '```Renku-User-FullName: Rok Roskar```'
                ),
                'in': 'header',
                'type': 'string',
                'required': True
            },
            'Renku-User-Email': {
                'description': (
                    'Used for commit author signature. '
                    'For example: '
                    '```Renku-User-Email: dev@renkulab.io```'
                ),
                'in': 'header',
                'type': 'string',
                'required': True
            },
        },
        tags=list(tags),
    )


def handle_common_except(f):
    """Handle common exceptions."""
    # noqa
    @wraps(f)
    def dec(*args, **kwargs):
        """Decorated function."""

        @handle_base_except
        @handle_validation_except
        @handle_renku_except
        @handle_git_except
        def _wrapped(*args_, **kwargs_):
            return f(*args_, **kwargs_)

        return _wrapped(*args, **kwargs)

    return dec



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represents a workflow template."""

import uuid

from renku.core.models import jsonld as jsonld


@jsonld.s(
    type=[
        'renku:IOStream',
    ],
    context={
        'renku': 'https://swissdatasciencecenter.github.io/renku-ontology#',
        'prov': 'http://www.w3.org/ns/prov#',
    },
    cmp=False,
)
class MappedIOStream(object):
    """Represents an IO stream (stdin, stdout, stderr)."""

    _id = jsonld.ib(context='@id', kw_only=True)
    _label = jsonld.ib(default=None, context='rdfs:label', kw_only=True)

    STREAMS = ['stdin', 'stdout', 'stderr']

    stream_type = jsonld.ib(
        context={
            '@id': 'renku:streamType',
            '@type': 'http://www.w3.org/2001/XMLSchema#string',
        },
        type=str,
        kw_only=True,
    )

    @_id.default
    def default_id(self):
        """Set default id."""
        # TODO: make bnode ids nicer once this issue is in a release:
        # https://github.com/RDFLib/rdflib/issues/888
        # right now it's limited to a-zA-Z0-9 (-_ will work once it's fixed)
        return '_:MappedIOStream-{}'.format(str(uuid.uuid4())).replace('-', '')

    def default_label(self):
        """Set default label."""
        return 'Stream mapping for stream "{}"'.format(self.stream_type)

    def __attrs_post_init__(self):
        """Post-init hook."""
        if not self._label:
            self._label = self.default_label()


@jsonld.s(
    type=[
        'renku:CommandParameter',
    ],
    context={
        'renku': 'https://swissdatasciencecenter.github.io/renku-ontology#',
        'prov': 'http://www.w3.org/ns/prov#',
    },
    cmp=False,
)
class CommandParameter(object):
    """Represents a parameter for an execution template."""

    _id = jsonld.ib(default=None, context='@id', kw_only=True)
    _label = jsonld.ib(default=None, context='rdfs:label', kw_only=True)

    position = jsonld.ib(
        default=None,
        context={
            '@id': 'renku:position',
            '@type': 'http://www.w3.org/2001/XMLSchema#integer',
        },
        type=int,
        kw_only=True,
    )

    prefix = jsonld.ib(
        default=None,
        context={
            '@id': 'renku:prefix',
            '@type': 'http://www.w3.org/2001/XMLSchema#string',
        },
        type=str,
        kw_only=True,
    )

    @property
    def sanitized_id(self):
        """Return ``_id`` sanitized for use in non-jsonld contexts."""
        return self._id.split(':', 1)[1].replace('-', '_')


@jsonld.s(
    type=[
        'renku:CommandArgument',
    ],
    context={
        'renku': 'https://swissdatasciencecenter.github.io/renku-ontology#',
        'prov': 'http://www.w3.org/ns/prov#',
    },
    cmp=False,
)
class CommandArgument(CommandParameter):
    """An argument to a command that is neither input nor output."""

    value = jsonld.ib(
        default=None,
        context={
            '@id': 'renku:value',
            '@type': 'http://www.w3.org/2001/XMLSchema#string',
        },
        type=str,
        kw_only=True,
    )

    def default_id(self):
        """Set default id."""
        return '_:CommandArgument-{}'.format(str(uuid.uuid4())
                                             ).replace('-', '')

    def default_label(self):
        """Set default label."""
        return 'Command Argument "{}"'.format(self.value)

    def to_argv(self):
        """String representation (sames as cmd argument)."""
        if self.prefix:
            if self.prefix.endswith(' '):
                return [self.prefix[:-1], self.value]
            return ['{}{}'.format(self.prefix, self.value)]

        return [self.value]

    def __attrs_post_init__(self):
        """Post-init hook."""
        if not self._id:
            self._id = self.default_id()

        if not self._label:
            self._label = self.default_label()


@jsonld.s(
    type=[
        'renku:CommandInput',
    ],
    context={
        'renku': 'https://swissdatasciencecenter.github.io/renku-ontology#',
        'prov': 'http://www.w3.org/ns/prov#',
    },
    cmp=False,
)
class CommandInput(CommandParameter):
    """An input to a command."""

    consumes = jsonld.ib(
        context='renku:consumes',
        kw_only=True,
        type=[
            'renku.core.models.entities.Entity',
            'renku.core.models.entities.Collection'
        ]
    )

    mapped_to = jsonld.ib(
        default=None,
        context='prov:mappedTo',
        kw_only=True,
        type=MappedIOStream
    )

    def default_id(self):
        """Set default id."""
        return '_:CommandInput-{}'.format(str(uuid.uuid4())).replace('-', '')

    def default_label(self):
        """Set default label."""
        return 'Command Input "{}"'.format(self.consumes.path)

    def to_argv(self):
        """String representation (sames as cmd argument)."""
        if self.prefix:
            if self.prefix.endswith(' '):
                return [self.prefix[:-1], self.consumes.path]
            return ['{}{}'.format(self.prefix, self.consumes.path)]

        return [self.consumes.path]

    def to_stream_repr(self):
        """Input stream representation."""
        if not self.mapped_to:
            return ''

        return ' < {}'.format(self.consumes.path)

    def __attrs_post_init__(self):
        """Post-init hook."""
        if not self._id:
            self._id = self.default_id()

        if not self._label:
            self._label = self.default_label()


@jsonld.s(
    type=[
        'renku:CommandOutput',
    ],
    context={
        'renku': 'https://swissdatasciencecenter.github.io/renku-ontology#',
        'prov': 'http://www.w3.org/ns/prov#',
    },
    cmp=False,
)
class CommandOutput(CommandParameter):
    """An output of a command."""

    create_folder = jsonld.ib(
        default=False, context='renku:createFolder', kw_only=True, type=bool
    )

    produces = jsonld.ib(
        context='renku:produces',
        kw_only=True,
        type=[
            'renku.core.models.entities.Entity',
            'renku.core.models.entities.Collection'
        ]
    )

    mapped_to = jsonld.ib(
        default=None,
        context='prov:mappedTo',
        kw_only=True,
        type=MappedIOStream
    )

    def default_id(self):
        """Set default id."""
        return '_:CommandOutput-{}'.format(str(uuid.uuid4())).replace('-', '')

    def default_label(self):
        """Set default label."""
        return 'Command Output "{}"'.format(self.produces.path)

    def to_argv(self):
        """String representation (sames as cmd argument)."""
        if self.prefix:
            if self.prefix.endswith(' '):
                return [self.prefix[:-1], self.produces.path]
            return ['{}{}'.format(self.prefix, self.produces.path)]

        return [self.produces.path]

    def to_stream_repr(self):
        """Input stream representation."""
        if not self.mapped_to:
            return ''

        if self.mapped_to.stream_type == 'stdout':
            return ' > {}'.format(self.produces.path)

        return ' 2> {}'.format(self.produces.path)

    def __attrs_post_init__(self):
        """Post-init hook."""
        if not self._id:
            self._id = self.default_id()

        if not self._label:
            self._label = self.default_label()



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service rpc serializers."""
from renku.service.serializers.datasets import DatasetListResponseRPC
from renku.service.views import error_response, result_response


def test_error_response(svc_client):
    """Test error response utility."""
    err_code = 0
    err_reason = 'test error'
    response = error_response(err_code, err_reason).json

    assert response
    assert {'error'} == set(response.keys())
    assert {'code', 'reason'} == set(response['error'].keys())

    assert err_code == response['error']['code']
    assert err_reason == response['error']['reason']


def test_result_response(svc_client):
    """Test result response utility."""
    ctx = {'datasets': [{'name': 'my-dataset'}]}
    response = result_response(DatasetListResponseRPC(), ctx).json

    assert response
    assert {'result'} == set(response.keys())
    assert {'datasets'} == set(response['result'].keys())

    expected = ctx['datasets'][0]
    received = response['result']['datasets'][0]

    assert expected['name'] == received['name']


def test_result_response_with_none(svc_client):
    """Test result response with None value."""
    response = result_response(DatasetListResponseRPC(), None).json

    assert response
    assert {'result'} == set(response.keys())
    assert response['result'] is None


def test_result_response_with_empty_dict(svc_client):
    """Test result response with empty value."""
    response = result_response(DatasetListResponseRPC(), {}).json

    assert response
    assert {'result'} == set(response.keys())
    assert {} == response['result']


def test_result_response_with_empty_tuple(svc_client):
    """Test result response with empty value."""
    response = result_response(DatasetListResponseRPC(), ()).json

    assert response
    assert {'result'} == set(response.keys())
    assert {} == response['result']



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test behavior of indirect inputs/outputs files list."""

from renku.core.models.entities import Collection, Entity


def test_indirect_inputs(cli, client):
    """Test indirect inputs that are programmatically created."""
    # Set up a script that creates indirect inputs
    cli('run', '--no-output', 'mkdir', 'foo')
    cli('run', '--no-output', 'mkdir', '.renku/tmp')
    cli('run', 'touch', 'foo/bar')
    cli('run', 'touch', 'baz')
    cli('run', 'touch', 'qux')
    cli(
        'run', 'sh', '-c',
        'echo "echo foo > .renku/tmp/inputs.txt" > script.sh'
    )
    cli(
        'run', 'sh', '-c',
        'echo "echo baz >> .renku/tmp/inputs.txt" >> script.sh'
    )
    cli(
        'run', 'sh', '-c',
        'echo "echo qux > .renku/tmp/outputs.txt" >> script.sh'
    )
    exit_code, cwl = cli('run', 'sh', '-c', 'sh script.sh')
    assert 0 == exit_code
    plan = cwl.association.plan
    assert 2 == len(plan.inputs)
    assert 1 == len(plan.arguments)
    plan.inputs.sort(key=lambda e: e.consumes.path)
    assert 'baz' == str(plan.inputs[0].consumes.path)
    assert isinstance(plan.inputs[0].consumes, Entity)
    assert plan.inputs[0].position is None
    assert 'foo' == str(plan.inputs[1].consumes.path)
    assert isinstance(plan.inputs[1].consumes, Collection)
    assert plan.inputs[1].position is None

    assert 1 == len(plan.outputs)
    assert 'qux' == plan.outputs[0].produces.path



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache views."""
import os
import shutil
from pathlib import Path

import patoolib
from flask import Blueprint, jsonify, request
from flask_apispec import marshal_with, use_kwargs
from git import Repo
from marshmallow import EXCLUDE
from patoolib.util import PatoolError

from renku.core.commands.clone import project_clone
from renku.core.commands.migrate import migrations_check
from renku.core.commands.save import repo_sync
from renku.core.utils.contexts import chdir
from renku.service.config import CACHE_UPLOADS_PATH, \
    INVALID_PARAMS_ERROR_CODE, SERVICE_PREFIX, SUPPORTED_ARCHIVES
from renku.service.jobs.contexts import enqueue_retry
from renku.service.jobs.project import execute_migration, migrate_job
from renku.service.jobs.queues import MIGRATIONS_JOB_QUEUE
from renku.service.serializers.cache import FileListResponseRPC, \
    FileUploadRequest, FileUploadResponseRPC, ProjectCloneContext, \
    ProjectCloneRequest, ProjectCloneResponseRPC, ProjectListResponseRPC, \
    ProjectMigrateAsyncResponseRPC, ProjectMigrateRequest, \
    ProjectMigrateResponseRPC, ProjectMigrationCheckResponseRPC, \
    extract_file
from renku.service.utils import make_project_path
from renku.service.views import result_response
from renku.service.views.decorators import accepts_json, \
    handle_common_except, header_doc, requires_cache, requires_identity

CACHE_BLUEPRINT_TAG = 'cache'
cache_blueprint = Blueprint('cache', __name__, url_prefix=SERVICE_PREFIX)


@marshal_with(FileListResponseRPC)
@header_doc(description='List uploaded files.', tags=(CACHE_BLUEPRINT_TAG, ))
@cache_blueprint.route(
    '/cache.files_list',
    methods=['GET'],
    provide_automatic_options=False,
)
@handle_common_except
@requires_cache
@requires_identity
def list_uploaded_files_view(user, cache):
    """List uploaded files ready to be added to projects."""
    user = cache.ensure_user(user)

    files = [f for f in cache.get_files(user) if f.exists()]

    response = {
        'files':
            sorted(files, key=lambda rec: (rec.is_dir, rec.relative_path))
    }

    return result_response(FileListResponseRPC(), response)


@use_kwargs(FileUploadRequest)
@marshal_with(FileUploadResponseRPC)
@header_doc(
    description='Upload file or archive of files.',
    tags=(CACHE_BLUEPRINT_TAG, ),
)
@cache_blueprint.route(
    '/cache.files_upload',
    methods=['POST'],
    provide_automatic_options=False,
)
@handle_common_except
@requires_cache
@requires_identity
def upload_file_view(user, cache):
    """Upload file or archive of files."""
    user = cache.ensure_user(user)
    file = extract_file(request)

    response_builder = {
        'file_name': file.filename,
        'content_type': file.content_type,
        'is_archive': file.content_type in SUPPORTED_ARCHIVES
    }
    response_builder.update(FileUploadRequest().load(request.args))

    user_cache_dir = CACHE_UPLOADS_PATH / user.user_id
    user_cache_dir.mkdir(exist_ok=True)

    file_path = user_cache_dir / file.filename
    if file_path.exists():
        if response_builder.get('override_existing', False):
            file_path.unlink()
        else:
            return jsonify(
                error={
                    'code': INVALID_PARAMS_ERROR_CODE,
                    'reason': 'file exists',
                }
            )

    file.save(str(file_path))

    files = []
    if response_builder['unpack_archive'] and response_builder['is_archive']:
        unpack_dir = '{0}.unpacked'.format(file_path.name)
        temp_dir = file_path.parent / Path(unpack_dir)
        if temp_dir.exists():
            shutil.rmtree(str(temp_dir))
        temp_dir.mkdir(exist_ok=True)

        try:
            patoolib.extract_archive(str(file_path), outdir=str(temp_dir))
        except PatoolError:
            return jsonify(
                error={
                    'code': INVALID_PARAMS_ERROR_CODE,
                    'reason': 'unable to unpack archive'
                }
            )

        for file_ in temp_dir.glob('**/*'):
            relative_path = file_.relative_to(
                CACHE_UPLOADS_PATH / user.user_id
            )

            file_obj = {
                'file_name': file_.name,
                'file_size': os.stat(str(file_path)).st_size,
                'relative_path': str(relative_path),
                'is_dir': relative_path.is_dir(),
            }

            files.append(file_obj)

    else:
        relative_path = file_path.relative_to(
            CACHE_UPLOADS_PATH / user.user_id
        )

        response_builder['file_size'] = os.stat(str(file_path)).st_size
        response_builder['relative_path'] = str(relative_path)
        response_builder['is_dir'] = relative_path.is_dir()

        files.append(response_builder)

    files = cache.set_files(user, files)
    return result_response(FileUploadResponseRPC(), {'files': files})


@requires_cache
def _project_clone(cache, user_data, project_data):
    """Clones the project for a given user."""
    local_path = make_project_path(user_data, project_data)
    user = cache.ensure_user(user_data)

    if local_path.exists():
        shutil.rmtree(str(local_path))

        for project in cache.get_projects(user):
            if project.git_url == project_data['git_url']:
                project.delete()

    local_path.mkdir(parents=True, exist_ok=True)
    project_clone(
        project_data['url_with_auth'],
        local_path,
        depth=project_data['depth'] if project_data['depth'] != 0 else None,
        raise_git_except=True,
        config={
            'user.name': project_data['fullname'],
            'user.email': project_data['email'],
        },
        checkout_rev=project_data['ref']
    )

    project = cache.make_project(user, project_data)
    return project


@use_kwargs(ProjectCloneRequest)
@marshal_with(ProjectCloneResponseRPC)
@header_doc(
    'Clone a remote project. If the project is cached already, '
    'new clone operation will override the old cache state.',
    tags=(CACHE_BLUEPRINT_TAG, )
)
@cache_blueprint.route(
    '/cache.project_clone',
    methods=['POST'],
    provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_identity
def project_clone_view(user_data):
    """Clone a remote repository."""
    project_data = ProjectCloneContext().load({
        **user_data,
        **request.json
    },
                                              unknown=EXCLUDE)
    project = _project_clone(user_data, project_data)

    return result_response(ProjectCloneResponseRPC(), project)


@marshal_with(ProjectListResponseRPC)
@header_doc(
    'List cached projects.',
    tags=(CACHE_BLUEPRINT_TAG, ),
)
@cache_blueprint.route(
    '/cache.project_list',
    methods=['GET'],
    provide_automatic_options=False,
)
@handle_common_except
@requires_cache
@requires_identity
def list_projects_view(user, cache):
    """List cached projects."""
    projects = [
        project for project in cache.get_projects(cache.ensure_user(user))
        if project.abs_path.exists()
    ]

    return result_response(ProjectListResponseRPC(), {'projects': projects})


@use_kwargs(ProjectMigrateRequest)
@marshal_with(ProjectMigrateResponseRPC)
@header_doc(
    'Migrate project to the latest version.',
    tags=(CACHE_BLUEPRINT_TAG, ),
)
@cache_blueprint.route(
    '/cache.migrate',
    methods=['POST'],
    provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def migrate_project_view(user_data, cache):
    """Migrate specified project."""
    ctx = ProjectMigrateRequest().load(request.json)
    user = cache.ensure_user(user_data)
    project = cache.get_project(user, ctx['project_id'])

    if ctx.get('is_delayed', False):
        job = cache.make_job(user, locked=project.project_id)

        with enqueue_retry(MIGRATIONS_JOB_QUEUE) as queue:
            queue.enqueue(
                migrate_job,
                user_data,
                project.project_id,
                job.job_id,
            )
        return result_response(ProjectMigrateAsyncResponseRPC(), job)

    messages, was_migrated = execute_migration(project)
    response = {'messages': messages, 'was_migrated': was_migrated}

    if was_migrated:
        _, response['remote_branch'] = repo_sync(
            Repo(project.abs_path), remote='origin'
        )

    return result_response(ProjectMigrateResponseRPC(), response)


@use_kwargs(ProjectMigrateRequest)
@marshal_with(ProjectMigrationCheckResponseRPC)
@header_doc(
    'Check if project requires migration.',
    tags=(CACHE_BLUEPRINT_TAG, ),
)
@cache_blueprint.route(
    '/cache.migrations_check',
    methods=['GET'],
    provide_automatic_options=False,
)
@handle_common_except
@requires_cache
@requires_identity
@accepts_json
def migration_check_project_view(user_data, cache):
    """Migrate specified project."""
    user = cache.ensure_user(user_data)
    project = cache.get_project(user, request.args['project_id'])

    with chdir(project.abs_path):
        migration_required, project_supported = migrations_check()

    return result_response(
        ProjectMigrationCheckResponseRPC(), {
            'migration_required': migration_required,
            'project_supported': project_supported
        }
    )



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache job related models."""
from walrus import DateTimeField, JSONField, Model, SetField, TextField

from renku.service.cache.base import BaseCache

# User job states
USER_JOB_STATE_ENQUEUED = 'ENQUEUED'
USER_JOB_STATE_IN_PROGRESS = 'IN_PROGRESS'
USER_JOB_STATE_COMPLETED = 'COMPLETED'
USER_JOB_STATE_FAILED = 'FAILED'


class Job(Model):
    """Job cache model."""

    __database__ = BaseCache.model_db

    created_at = DateTimeField()
    updated_at = DateTimeField()

    job_id = TextField(primary_key=True, index=True)
    user_id = TextField(index=True)

    state = TextField()
    extras = JSONField()

    locked = SetField()

    def in_progress(self):
        """Mark job in progress."""
        self.state = USER_JOB_STATE_IN_PROGRESS
        self.save()

    def complete(self):
        """Mark job as completed."""
        self.state = USER_JOB_STATE_COMPLETED
        self.save()

    def fail_job(self, error):
        """Mark job as failed."""
        self.state = USER_JOB_STATE_FAILED

        if not self.extras:
            self.extras = {}

        self.extras['error'] = error
        self.save()

    def update_extras(self, key, value):
        """Update extras field."""
        if not self.extras:
            self.extras = {key: value}
        else:
            self.extras[key] = value



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Serialization tests for renku models."""

import datetime
from urllib.parse import urljoin

import yaml

from renku.core.management.client import LocalClient
from renku.core.models.datasets import Dataset
from renku.core.utils.uuid import is_uuid


def test_dataset_serialization(client, dataset, data_file):
    """Test Dataset serialization."""

    def load_dataset(name):
        with open(str(client.get_dataset_path(name))) as f:
            return yaml.safe_load(f)

    d_dict = load_dataset('dataset')

    expected_fields = [
        '@id', 'http://www.w3.org/2000/01/rdf-schema#label',
        'http://schema.org/isPartOf', 'http://schema.org/dateCreated',
        'http://schema.org/creator', 'http://schema.org/datePublished',
        'http://schema.org/description', 'http://schema.org/hasPart',
        'http://schema.org/identifier', 'http://schema.org/inLanguage',
        'http://schema.org/keywords', 'http://schema.org/license',
        'http://schema.org/name', 'http://www.w3.org/ns/prov#atLocation',
        'http://schema.org/url', 'http://schema.org/version'
    ]
    for field in expected_fields:
        assert field in d_dict

    assert not d_dict['http://schema.org/hasPart']
    client.add_data_to_dataset(dataset, [str(data_file)])
    dataset.to_yaml()
    d_dict = load_dataset('dataset')
    assert d_dict['http://schema.org/hasPart']


def test_dataset_deserialization(client, dataset):
    """Test Dataset deserialization."""
    dataset_ = Dataset.from_yaml(
        client.get_dataset_path('dataset'), client=client
    )

    dataset_types = {
        'date_created': [datetime.datetime],
        'creators': [list],
        'description': [str, type(None)],
        'files': [list],
        'identifier': [str],
        'keywords': [list],
    }

    for attribute, type_ in dataset_types.items():
        assert type(dataset_.__getattribute__(attribute)) in type_

    creator_types = {'email': str, '_id': str, 'name': str, 'affiliation': str}

    creator = dataset.creators[0]

    for attribute, type_ in creator_types.items():
        assert type(getattr(creator, attribute)) is type_


def test_dataset_files_empty_metadata(dataset_metadata):
    """Check parsing metadata of dataset files with empty filename."""
    dataset = Dataset.from_jsonld(
        dataset_metadata,
        client=LocalClient('.'),
    )
    files = [file.filename for file in dataset.files if not file.filename]

    if files:
        assert None in files


def test_uuid_migration(dataset_metadata, client):
    """Test migration of id with UUID."""
    dataset = Dataset.from_jsonld(dataset_metadata, client=client)

    assert is_uuid(dataset.identifier)
    assert urljoin(
        'https://localhost/datasets/', dataset.identifier
    ) == dataset._id


def test_dataset_creator_email(dataset_metadata):
    """Check that creators without an email are assigned a blank node."""
    # modify the dataset metadata to change the creator
    dataset = Dataset.from_jsonld(
        dataset_metadata,
        client=LocalClient('.'),
    )

    dataset.creators[0]._id = 'mailto:None'
    dataset_broken = Dataset.from_jsonld(
        dataset.asjsonld(), client=LocalClient('.')
    )
    assert 'mailto:None' not in dataset_broken.creators[0]._id


def test_calamus(client, dataset_metadata_before_calamus):
    """Check Calamus loads project correctly."""
    dataset = Dataset.from_jsonld(
        dataset_metadata_before_calamus, client=LocalClient('.')
    )
    assert 'Open Source at Harvard' == dataset.title
    assert '51db02ad-3cba-47e2-84d0-5ee5914bd654' == dataset.identifier
    assert '51db02ad-3cba-47e2-84d0-5ee5914bd654' == dataset._label
    assert 'Harvard University' == dataset.creators[0].affiliation
    assert 'Durbin, Philip' == dataset.creators[0].name
    assert 'Durbin, Philip' == dataset.creators[0].label
    assert dataset.date_created is None
    assert '2019-07-03T00:00:00' == dataset.date_published.isoformat('T')
    assert 'The tabular file contains information' in dataset.description
    assert 'https://doi.org/10.7910/DVN/TJCLKP' == dataset.same_as.url
    assert '3' == dataset.tags[0].name
    assert 'Tag 3 created by renku import' == dataset.tags[0].description
    assert isinstance(dataset.license, dict)
    assert (
        'https://creativecommons.org/publicdomain/zero/1.0/' in str(
            dataset.license
        )
    )

    file_ = dataset.find_file('data/dataverse/IQSS-UNF.json')
    assert (
        'https://dataverse.harvard.edu/api/access/datafile/3371500' ==
        file_.url
    )
    assert '2020-06-15T08:37:04.571573+00:00' == file_.added.isoformat('T')
    assert file_.based_on is None

    file_ = dataset.find_file('data/dataverse/git/index.ipynb')
    assert (
        'https://github.com/SwissDataScienceCenter/r10e-ds-py.git' ==
        file_.based_on.url
    )
    assert (
        'notebooks/index.ipynb@f98325d81c700f4b86ee05c2154e94d43ca068b8' ==
        file_.based_on._label
    )
    assert file_.based_on.based_on is None
    assert (
        'https://github.com/SwissDataScienceCenter/r10e-ds-py.git' == file_.url
    )

    file_ = dataset.find_file('data/dataverse/external/data.txt')
    assert file_.external is True
    assert 'file://../../../../tmp/data.txt' == file_.url

    file_ = dataset.find_file('data/dataverse/local/result.csv')
    assert file_.external is False
    assert 'file://../../../../tmp/result.csv' == file_.url



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Move or rename a file, a directory, or a symlink.

Moving a file that belongs to a dataset will update its metadata. It also
will attempt to update tracking information for files stored in an external
storage (using Git LFS). Finally it makes sure that all relative symlinks work
after the move.
"""

import os
from pathlib import Path
from subprocess import run

import click

from renku.core.commands.client import pass_local_client
from renku.core.commands.echo import INFO, WARNING, progressbar


@click.command(name='mv')
@click.argument('sources', type=click.Path(exists=True), nargs=-1)
@click.argument('destination', type=click.Path(), nargs=1)
@pass_local_client(
    clean=True,
    requires_migration=True,
    commit=True,
)
@click.pass_context
def move(ctx, client, sources, destination):
    """Move files and check repository for potential problems."""
    from renku.core.management.git import _expand_directories

    dst = Path(destination)

    def fmt_path(path):
        """Format path as relative to the client path."""
        return str(Path(path).absolute().relative_to(client.path))

    files = {
        fmt_path(source): fmt_path(file_or_dir)
        for file_or_dir in sources
        for source in _expand_directories((file_or_dir, ))
    }

    def fmt_dst(path):
        """Build a destination path for a source path."""
        return str(dst / os.path.relpath(path, start=files[path]))

    destinations = {source: fmt_dst(source) for source in files}

    # 1. Check .gitignore.
    ignored = client.find_ignored_paths(*destinations.values())
    if ignored:
        click.echo(WARNING + 'Renamed files match .gitignore.\n')
        if click.confirm(
            'Do you want to edit ".gitignore" now?', default=False
        ):
            click.edit(filename=str(client.path / '.gitignore'))

    # 2. Update dataset metadata files.
    with progressbar(
        client.datasets.items(),
        item_show_func=lambda item: str(item[1].short_id) if item else '',
        label='Updating dataset metadata',
        width=0,
    ) as bar:
        for (path, dataset) in bar:
            renames = {}

            for file_ in dataset.files:
                filepath = fmt_path(file_.path)

                if filepath in files:
                    renames[file_.path] = destinations[filepath]

            if renames:
                dataset = dataset.rename_files(
                    lambda key: renames.get(key, key)
                )

                dataset.to_yaml()

    # 3. Manage .gitattributes for external storage.
    tracked = tuple()
    if client.check_external_storage():
        tracked = tuple(
            path for path, attr in client.find_attr(*files).items()
            if attr.get('filter') == 'lfs'
        )
        client.untrack_paths_from_storage(*tracked)

        if client.find_attr(*tracked):
            click.echo(WARNING + 'There are custom .gitattributes.\n')
            if click.confirm(
                'Do you want to edit ".gitattributes" now?', default=False
            ):
                click.edit(filename=str(client.path / '.gitattributes'))

        if tracked:
            lfs_paths = client.track_paths_in_storage(
                *(destinations[path] for path in tracked)
            )
            show_message = client.get_value('renku', 'show_lfs_message')
            if lfs_paths and (show_message is None or show_message == 'True'):
                click.echo(
                    INFO + 'Adding these files to Git LFS:\n' +
                    '\t{}'.format('\n\t'.join(lfs_paths)) +
                    '\nTo disable this message in the future, run:' +
                    '\n\trenku config show_lfs_message False'
                )

    # 4. Handle symlinks.
    dst.parent.mkdir(parents=True, exist_ok=True)

    for source, target in destinations.items():
        src = Path(source)
        if src.is_symlink():
            Path(target).parent.mkdir(parents=True, exist_ok=True)
            Path(target).symlink_to(
                os.path.relpath(
                    str(src.resolve()), start=os.path.dirname(target)
                )
            )
            src.unlink()
            del files[source]

    # Finally move the files.
    final_sources = list(set(files.values()))
    if final_sources:
        run(['git', 'mv'] + final_sources + [destination], check=True)



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client for handling a local repository."""
import os
import shutil
import subprocess
import uuid
from collections import defaultdict
from contextlib import contextmanager
from subprocess import check_output

import attr
import filelock
import yaml
from jinja2 import Template
from werkzeug.utils import cached_property, secure_filename

from renku.core import errors
from renku.core.compat import Path
from renku.core.management.config import RENKU_HOME
from renku.core.models.locals import with_reference
from renku.core.models.projects import Project
from renku.core.models.refs import LinkReference

from .git import GitCore

DEFAULT_DATA_DIR = 'data'


def default_path():
    """Return default repository path."""
    from git import InvalidGitRepositoryError

    from renku.core.commands.git import get_git_home
    try:
        return get_git_home()
    except InvalidGitRepositoryError:
        return '.'


def path_converter(path):
    """Converter for path in PathMixin."""
    return Path(path).resolve()


@attr.s
class PathMixin:
    """Define a default path attribute."""

    path = attr.ib(
        default=default_path,
        converter=path_converter,
    )

    @path.validator
    def _check_path(self, _, value):
        """Check the path exists and it is a directory."""
        if not (value.exists() and value.is_dir()):
            raise ValueError('Define an existing directory.')


@attr.s
class RepositoryApiMixin(GitCore):
    """Client for handling a local repository."""

    renku_home = attr.ib(default=RENKU_HOME)
    """Define a name of the Renku folder (default: ``.renku``)."""

    renku_path = attr.ib(init=False)
    """Store a ``Path`` instance of the Renku folder."""

    parent = attr.ib(default=None)
    """Store a pointer to the parent repository."""

    data_dir = attr.ib(
        default=DEFAULT_DATA_DIR,
        kw_only=True,
        converter=lambda value: str(value) if value else DEFAULT_DATA_DIR
    )
    """Define a name of the folder for storing datasets."""

    METADATA = 'metadata.yml'
    """Default name of Renku config file."""

    LOCK_SUFFIX = '.lock'
    """Default suffix for Renku lock file."""

    WORKFLOW = 'workflow'
    """Directory for storing workflow in Renku."""

    ACTIVITY_INDEX = 'activity_index.yaml'
    """Caches activities that generated a path."""

    RENKU_PROTECTED_PATHS = [
        '\\.renku/.*', 'Dockerfile', '\\.dockerignore', '\\.gitignore',
        '\\.gitattributes', '\\.gitlab-ci\\.yml', 'environment\\.yml',
        'requirements\\.txt'
    ]

    _commit_activity_cache = {}

    _activity_index = None

    _remote_cache = {}

    def __attrs_post_init__(self):
        """Initialize computed attributes."""
        #: Configure Renku path.
        path = Path(self.renku_home)
        if not path.is_absolute():
            path = self.path / path

        path.relative_to(path)
        self.renku_path = path

        data_dir = self.get_value(
            'renku', self.DATA_DIR_CONFIG_KEY, local_only=True
        )
        self.data_dir = data_dir or self.data_dir

        self._subclients = {}

        self._project = None

        super().__attrs_post_init__()

        # initialize submodules
        if self.repo:
            try:
                check_output([
                    'git', 'submodule', 'update', '--init', '--recursive'
                ],
                             cwd=str(self.path))
            except subprocess.CalledProcessError:
                pass

    @property
    def lock(self):
        """Create a Renku config lock."""
        return filelock.FileLock(
            str(self.renku_path.with_suffix(self.LOCK_SUFFIX)),
            timeout=0,
        )

    @property
    def renku_metadata_path(self):
        """Return a ``Path`` instance of Renku metadata file."""
        return self.renku_path.joinpath(self.METADATA)

    @property
    def workflow_path(self):
        """Return a ``Path`` instance of the workflow folder."""
        return self.renku_path / self.WORKFLOW

    @property
    def activity_index_path(self):
        """Path to the activity filepath cache."""
        return self.renku_path / self.ACTIVITY_INDEX

    @cached_property
    def cwl_prefix(self):
        """Return a CWL prefix."""
        self.workflow_path.mkdir(parents=True, exist_ok=True)  # for Python 3.5
        return str(self.workflow_path.resolve().relative_to(self.path))

    @property
    def project(self):
        """Return the Project instance."""
        if self.renku_metadata_path.exists() and self._project is None:
            self._project = Project.from_yaml(
                self.renku_metadata_path, client=self
            )
        return self._project

    @property
    def remote(self, remote_name='origin'):
        """Return host, owner and name of the remote if it exists."""
        from renku.core.models.git import GitURL

        original_remote_name = remote_name

        if original_remote_name in self._remote_cache:
            return self._remote_cache[original_remote_name]

        host = owner = name = None
        try:
            remote_branch = \
                self.repo.head.reference.tracking_branch()
            if remote_branch is not None:
                remote_name = remote_branch.remote_name
        except TypeError:
            pass

        try:
            url = GitURL.parse(self.repo.remotes[remote_name].url)

            # Remove gitlab. unless running on gitlab.com.
            hostname_parts = url.hostname.split('.')
            if len(hostname_parts) > 2 and hostname_parts[0] == 'gitlab':
                hostname_parts = hostname_parts[1:]
            url = attr.evolve(url, hostname='.'.join(hostname_parts))
        except IndexError:
            url = None

        if url:
            host = url.hostname
            owner = url.owner
            name = url.name

        remote = {'host': host, 'owner': owner, 'name': name}
        self._remote_cache[original_remote_name] = remote

        return remote

    def is_project_set(self):
        """Return if project is set for the client."""
        return self._project is not None

    def process_commit(self, commit=None, path=None):
        """Build an :class:`~renku.core.models.provenance.activities.Activity`.

        :param commit: Commit to process. (default: ``HEAD``)
        :param path: Process a specific CWL file.
        """
        from renku.core.models.provenance.activities import Activity

        commit = commit or self.repo.head.commit
        if len(commit.parents) > 1:
            return Activity(commit=commit, client=self)

        if path is None:
            for file_ in commit.stats.files.keys():
                # Find a process (CommandLineTool or Workflow)
                if self.is_workflow(file_):
                    if path is not None:
                        # Regular activity since it edits multiple CWL files
                        return Activity(commit=commit, client=self)

                    path = file_

        if not path:
            # search for activities a file could have been a part of
            activities = self.activities_for_paths(
                commit.stats.files.keys(), file_commit=commit, revision='HEAD'
            )
            if len(activities) > 1:
                raise errors.CommitProcessingError(
                    'Found multiple activities that produced the same entity '
                    'at commit {}'.format(commit)
                )
            if activities:
                return activities[0]

        if path:
            data = (commit.tree / path).data_stream.read()
            process = Activity.from_jsonld(
                yaml.safe_load(data),
                client=self,
                commit=commit,
                __reference__=path
            )

            return process

        return Activity(commit=commit, client=self)

    def is_workflow(self, path):
        """Check if the path is a valid CWL file."""
        return path.startswith(self.cwl_prefix) and path.endswith('.yaml')

    def find_previous_commit(
        self, paths, revision='HEAD', return_first=False, full=False
    ):
        """Return a previous commit for a given path starting from ``revision``.

        :param revision: revision to start from, defaults to ``HEAD``
        :param return_first: show the first commit in the history
        :param full: return full history
        :raises KeyError: if path is not present in the given commit
        """
        kwargs = {}

        if full:
            kwargs['full_history'] = True

        if return_first:
            file_commits = list(
                self.repo.iter_commits(revision, paths=paths, **kwargs)
            )
        else:
            file_commits = list(
                self.repo.iter_commits(
                    revision, paths=paths, max_count=1, **kwargs
                )
            )

        if not file_commits:
            raise KeyError(
                'Could not find a file {0} in range {1}'.format(
                    paths, revision
                )
            )

        return file_commits[-1 if return_first else 0]

    @cached_property
    def workflow_names(self):
        """Return index of workflow names."""
        names = defaultdict(list)
        for ref in LinkReference.iter_items(self, common_path='workflows'):
            names[str(ref.reference.relative_to(self.path))].append(ref.name)
        return names

    @cached_property
    def submodules(self):
        """Return list of submodules it belongs to."""
        if self.parent:
            client, submodule = self.parent
            return client.submodules + [submodule.name]
        return []

    def subclients(self, parent_commit):
        """Return mapping from submodule to client."""
        if parent_commit in self._subclients:
            return self._subclients[parent_commit]

        try:
            from git import Submodule

            submodules = [
                submodule for submodule in
                Submodule.iter_items(self.repo, parent_commit=parent_commit)
            ]
        except (RuntimeError, ValueError):
            # There are no submodules associated with the given commit.
            submodules = []

        subclients = {}
        for submodule in submodules:
            subpath = (self.path / submodule.path).resolve()
            is_renku = subpath / Path(self.renku_home)

            if subpath.exists() and is_renku.exists():
                subclients[submodule] = self.__class__(
                    path=subpath,
                    parent=(self, submodule),
                )

        return subclients

    def resolve_in_submodules(self, commit, path):
        """Resolve filename in submodules."""
        original_path = self.path / path
        in_vendor = str(path).startswith('.renku/vendors')

        if original_path.is_symlink() or in_vendor:
            original_path = Path(
                os.path.realpath(os.path.abspath(str(original_path)))
            )

            for submodule, subclient in self.subclients(commit).items():
                if (Path(submodule.path) / Path('.git')).exists():

                    try:
                        subpath = original_path.relative_to(subclient.path)
                        return (
                            subclient,
                            subclient.find_previous_commit(
                                subpath, revision=submodule.hexsha
                            ),
                            subpath,
                        )
                    except ValueError:
                        pass

        return self, commit, path

    @contextmanager
    def with_commit(self, commit):
        """Yield the state of the repo at a specific commit."""
        current_branch = None
        current_commit = None

        try:
            current_branch = self.repo.active_branch
        except TypeError as e:
            # not on a branch, detached head
            if 'HEAD is a detached' in str(e):
                current_commit = self.repo.head.commit
            else:
                raise ValueError('Couldn\'t get active branch or commit', e)

        self.repo.git.checkout(commit)

        try:
            yield
        finally:
            if current_branch:
                self.repo.git.checkout(current_branch)
            elif current_commit:
                self.repo.git.checkout(current_commit)

    @contextmanager
    def with_metadata(self, read_only=False, name=None):
        """Yield an editable metadata object."""
        metadata_path = self.renku_metadata_path

        if metadata_path.exists():
            metadata = Project.from_yaml(metadata_path, client=self)
        else:
            metadata = Project(name=name, client=self)
            metadata.__reference__ = metadata_path

        yield metadata

        if not read_only:
            metadata.to_yaml()

    @contextmanager
    def with_workflow_storage(self):
        """Yield a workflow storage."""
        from renku.core.models.cwl.workflow import Workflow

        workflow = Workflow()
        yield workflow

        for step in workflow.steps:
            step_name = '{0}_{1}.yaml'.format(
                uuid.uuid4().hex,
                secure_filename('_'.join(step.run.baseCommand)),
            )

            workflow_path = self.workflow_path
            if not workflow_path.exists():
                workflow_path.mkdir()

            path = workflow_path / step_name

            with with_reference(path):
                run = step.run.generate_process_run(
                    client=self,
                    commit=self.repo.head.commit,
                    path=path,
                )
                run.to_yaml()
                self.add_to_activity_index(run)

    def init_repository(self, force=False, user=None):
        """Initialize an empty Renku repository."""
        from git import Repo

        from renku.core.models.provenance.agents import Person

        # verify if folder is empty
        if self.repo is not None and not force:
            raise errors.InvalidFileOperation(
                'Folder {0} already contains file. Use --force to overwrite'.
                format(self.repo.git_dir)
            )

        # initialize repo and set user data
        path = self.path.absolute()
        self.repo = Repo.init(str(path))
        if user:
            config_writer = self.repo.config_writer()
            for key, value in user.items():
                config_writer.set_value('user', key, value)
            config_writer.release()

        # verify if author information is available
        Person.from_git(self.repo)

    @property
    def path_activity_cache(self):
        """Cache of all activities and their generated paths."""
        # TODO: this is there for performance reasons. Remove once graph
        # is stored as a flat, append-only list (Should be graph query
        # in the future)
        if self._activity_index:
            return self._activity_index

        path = self.activity_index_path

        if path.exists():
            with path.open('r') as stream:
                self._activity_index = yaml.safe_load(stream)
        else:
            self._activity_index = {}

        return self._activity_index

    def add_to_activity_index(self, activity):
        """Add an activity and it's generations to the cache."""
        for g in activity.generated:
            if g.path not in self.path_activity_cache:
                self.path_activity_cache[g.path] = {}
            hexsha = g.commit.hexsha
            if hexsha not in self.path_activity_cache[g.path]:
                self.path_activity_cache[g.path][hexsha] = []

            if activity.path in self.path_activity_cache[g.path][hexsha]:
                continue

            self.path_activity_cache[g.path][g.commit.hexsha].append(
                activity.path
            )

        if self.path_activity_cache:
            with self.activity_index_path.open('w') as stream:
                yaml.dump(self.path_activity_cache, stream)

    def activities_for_paths(self, paths, file_commit=None, revision='HEAD'):
        """Get all activities involving a path."""
        from renku.core.models.provenance.activities import Activity

        result = set()

        for p in paths:
            if p not in self.path_activity_cache:
                parent_paths = [
                    a for a in self.path_activity_cache.keys()
                    if p.startswith(a)
                ]
                if not parent_paths:
                    continue
                matching_activities = [
                    a for k, v in self.path_activity_cache.items()
                    for ck, cv in v.items()
                    for a in cv if k in parent_paths and
                    (not file_commit or ck == file_commit.hexsha)
                ]
            else:
                matching = self.path_activity_cache[p]
                if file_commit:
                    if file_commit.hexsha not in matching:
                        continue
                    matching_activities = matching[file_commit.hexsha]
                else:
                    matching_activities = [
                        a for v in matching.values for a in v
                    ]

            for a in matching_activities:
                if (a, revision) in self._commit_activity_cache:
                    activity = self._commit_activity_cache[(a, revision)]
                else:
                    try:
                        commit = self.find_previous_commit(
                            a, revision=revision
                        )
                    except KeyError:
                        continue

                    activity = Activity.from_yaml(
                        self.path / a, client=self, commit=commit
                    )

                    self._commit_activity_cache[(a, revision)] = activity
                result.add(activity)

        return list(result)

    def import_from_template(self, template_path, metadata, force=False):
        """Render template files from a template directory."""
        for file in template_path.glob('**/*'):
            destination = self.path / file.relative_to(template_path)
            try:
                # TODO: notify about the expected variables - code stub:
                # with file.open() as fr:
                #     file_content = fr.read()
                #     # look for the required keys
                #     env = Environment()
                #     parsed = env.parse(file_content)
                #     variables = meta.find_undeclared_variables(parsed)

                # parse file and process it
                template = Template(file.read_text())
                rendered_content = template.render(metadata)
                destination = Path(Template(str(destination)).render(metadata))
                destination.write_text(rendered_content)
            except IsADirectoryError:
                destination.mkdir(parents=True, exist_ok=True)
            except TypeError:
                shutil.copy(file, destination)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Get and set Renku repository or global options."""
from renku.core import errors
from renku.core.management.config import CONFIG_LOCAL_PATH

from .client import pass_local_client


def _split_section_and_key(key):
    """Return a tuple with config section and key."""
    parts = key.split('.')
    if len(parts) > 1:
        return 'renku "{0}"'.format(parts[0]), '.'.join(parts[1:])
    return 'renku', key


@pass_local_client(
    clean=False,
    requires_migration=True,
    commit=True,
    commit_only=CONFIG_LOCAL_PATH,
    commit_empty=False
)
def update_config(
    client,
    key,
    *,
    value=None,
    remove=False,
    global_only=False,
    commit_message=None
):
    """Add, update, or remove configuration values."""
    section, section_key = _split_section_and_key(key)
    if remove:
        value = client.remove_value(
            section, section_key, global_only=global_only
        )
        if value is None:
            raise errors.ParameterError('Key "{}" not found.'.format(key))
    else:
        client.set_value(section, section_key, value, global_only=global_only)
        return value


@pass_local_client
def read_config(client, key, local_only, global_only):
    """Read configuration."""
    if key:
        section, section_key = _split_section_and_key(key)
        value = client.get_value(
            section,
            section_key,
            local_only=local_only,
            global_only=global_only
        )
        if value is None:
            raise errors.ParameterError('Key "{}" not found.'.format(key))
        return value

    return client.get_config(local_only=local_only, global_only=global_only)



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper utils for migrations."""


def migrate_types(data):
    """Fix data types."""
    type_mapping = {
        'dcterms:creator': ['prov:Person', 'schema:Person'],
        'schema:Person': ['prov:Person', 'schema:Person'],
        str(sorted(['foaf:Project', 'prov:Location'])): [
            'prov:Location', 'schema:Project'
        ],
        'schema:DigitalDocument': [
            'prov:Entity', 'schema:DigitalDocument', 'wfprov:Artifact'
        ]
    }

    def replace_types(data):
        for key, value in data.items():
            if key == '@type':
                if not isinstance(value, str):
                    value = str(sorted(value))
                new_type = type_mapping.get(value)
                if new_type:
                    data[key] = new_type
            elif isinstance(value, dict):
                replace_types(value)
            elif isinstance(value, (list, tuple, set)):
                for v in value:
                    if isinstance(v, dict):
                        replace_types(v)

    replace_types(data)

    return data



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``init`` command."""
import os
from pathlib import Path

import pytest

from renku.cli import cli
from renku.cli.init import create_template_sentence, parse_parameters
from renku.core import errors
from tests.core.commands.test_init import METADATA, TEMPLATE_ID, \
    TEMPLATE_INDEX, TEMPLATE_REF, TEMPLATE_URL
from tests.utils import raises

INIT = ['init', 'test-new-project', '--template-id', TEMPLATE_ID]
INIT_REMOTE = [
    '--template-source', TEMPLATE_URL, '--template-ref', TEMPLATE_REF
]
INIT_FORCE = ['--force']
INIT_INDEX = ['init', 'test-new-project-2', '--template-index', TEMPLATE_INDEX]
INIT_ID = ['--template-id', TEMPLATE_ID]
LIST_TEMPLATES = ['--list-templates']
PARAMETERS_CORRECT = ['--parameter', 'p1=v1', '--parameter', 'p2=v2']
PARAMETER_NO_EQUAL = ['--parameter', 'p3:v3']
PARAMETER_EARLY_EQUAL = ['--parameter', '=p4v3']
INPUT = len(set(METADATA.keys())) * '\n'


def test_parse_parameters():
    def clean_param(p):
        return [v for v in p if v != '--parameter']

    parsed = parse_parameters(None, None, clean_param(PARAMETERS_CORRECT))
    keys = parsed.keys()
    assert 2 == len(keys)
    assert 'p1' in keys
    assert 'p2' in keys
    assert 'v1' == parsed['p1']
    assert 'v2' == parsed['p2']
    with raises(errors.ParameterError):
        parsed = parse_parameters(
            None, None,
            clean_param(PARAMETERS_CORRECT) + clean_param(PARAMETER_NO_EQUAL)
        )
    with raises(errors.ParameterError):
        parsed = parse_parameters(
            None, None,
            clean_param(PARAMETERS_CORRECT) +
            clean_param(PARAMETER_EARLY_EQUAL)
        )


def test_template_selection_helpers():
    templates = [{
        'name': 'Template Python',
        'folder': 'folder_python',
        'description': 'Description Python'
    }, {
        'name': 'Template R',
        'folder': 'folder_R',
        'description': 'Description R',
        'variables': {
            'custom': 'random data'
        }
    }]
    instructions = 'Please choose a template by typing the index'
    sentence = create_template_sentence(templates)
    stripped_sentence = ' '.join(sentence.split())

    assert '1 folder_python' in stripped_sentence
    assert '2 folder_R' in stripped_sentence
    assert instructions not in stripped_sentence
    full_sentence = create_template_sentence(templates, instructions=True)
    assert instructions in full_sentence

    # with describe=True
    sentence = create_template_sentence(templates, describe=True)
    stripped_sentence = ' '.join(sentence.split())

    assert '1 folder_python Description Python' in stripped_sentence
    assert '2 folder_R custom: random data Description R' in stripped_sentence
    assert instructions not in stripped_sentence

    full_sentence = create_template_sentence(
        templates, describe=True, instructions=True
    )
    assert instructions in full_sentence


def test_list_templates(isolated_runner):
    """Test listing templates."""
    new_project = Path('test-new-project')
    assert not new_project.exists()
    result = isolated_runner.invoke(cli, INIT + LIST_TEMPLATES)
    assert 0 == result.exit_code
    assert not new_project.exists()
    assert TEMPLATE_ID in result.output


def test_init(isolated_runner):
    """Test project initialization from template."""
    # create the project
    new_project = Path('test-new-project')
    assert not new_project.exists()
    result = isolated_runner.invoke(cli, INIT, INPUT)
    assert 0 == result.exit_code
    assert new_project.exists()
    assert (new_project / '.renku').exists()
    assert (new_project / '.renku' / 'renku.ini').exists()
    assert (new_project / '.renku' / 'metadata.yml').exists()

    # try to re-create in the same folder
    result_re = isolated_runner.invoke(cli, INIT, INPUT)
    assert 0 != result_re.exit_code

    # force re-create in the same folder
    result_re = isolated_runner.invoke(cli, INIT + INIT_FORCE, INPUT)
    assert 0 == result.exit_code
    assert new_project.exists()
    assert (new_project / '.renku').exists()
    assert (new_project / '.renku' / 'renku.ini').exists()
    assert (new_project / '.renku' / 'metadata.yml').exists()

    # init using index instead of id
    new_project_2 = Path('test-new-project-2')
    result = isolated_runner.invoke(cli, INIT_INDEX, INPUT)
    assert 0 == result.exit_code
    assert new_project_2.exists()
    assert (new_project_2 / '.renku').exists()
    assert (new_project_2 / '.renku' / 'renku.ini').exists()
    assert (new_project_2 / '.renku' / 'metadata.yml').exists()

    # verify both init lead to the same result
    template_files = [
        f for f in new_project.glob('**/*') if '.git' not in str(f)
    ]
    for template_file in template_files:
        expected_file = new_project_2 / template_file.relative_to(new_project)
        assert expected_file.exists()

    # verify providing both index and id fails
    result = isolated_runner.invoke(
        cli, INIT_INDEX + INIT_ID + INIT_FORCE, INPUT
    )
    assert 2 == result.exit_code
    assert (
        'Use either --template-id or --template-index, not both'
    ) in result.output


def test_init_force_in_empty_dir(isolated_runner):
    """Run init --force in empty directory."""
    new_project = Path('test-new-project')
    assert not new_project.exists()
    result = isolated_runner.invoke(cli, INIT + INIT_FORCE, INPUT)
    assert 0 == result.exit_code


def test_init_force_in_dirty_dir(isolated_runner):
    """Run init --force in dirty directory."""
    new_project = Path('test-new-project')
    assert not new_project.exists()

    new_project.mkdir(parents=True)
    random_file = new_project / 'random_file.txt'
    with random_file.open('w') as dest:
        dest.writelines(['random text'])
    assert random_file.exists()

    result = isolated_runner.invoke(cli, INIT + INIT_FORCE, INPUT)
    assert random_file.exists()
    assert 0 == result.exit_code


def test_init_on_cloned_repo(isolated_runner, data_repository):
    """Run init --force in directory containing another repo."""
    new_project = Path('test-new-project')
    import shutil
    shutil.copytree(str(data_repository.working_dir), str(new_project))
    assert new_project.exists()

    # try to create in a dirty folder
    result = isolated_runner.invoke(cli, INIT, INPUT)
    assert 0 != result.exit_code

    # force re-create in the same folder
    result = isolated_runner.invoke(cli, INIT + INIT_FORCE, INPUT)
    assert 0 == result.exit_code
    assert new_project.exists()
    assert (new_project / '.renku').exists()
    assert (new_project / '.renku' / 'renku.ini').exists()
    assert (new_project / '.renku' / 'metadata.yml').exists()


@pytest.mark.integration
def test_init_remote(isolated_runner):
    """Test project initialization from a remote template."""
    # create the project
    new_project = Path('test-new-project')
    assert not new_project.exists()
    result = isolated_runner.invoke(cli, INIT + INIT_REMOTE, INPUT)
    assert 0 == result.exit_code
    assert new_project.exists()
    assert (new_project / '.renku').exists()
    assert (new_project / '.renku' / 'renku.ini').exists()
    assert (new_project / '.renku' / 'metadata.yml').exists()


def test_init_with_parameters(isolated_runner):
    """Test project initialization using custom metadata."""
    # create the project
    new_project = Path('test-new-project')
    assert not new_project.exists()
    result = isolated_runner.invoke(
        cli, INIT + PARAMETERS_CORRECT + PARAMETER_NO_EQUAL
    )
    assert 0 != result.exit_code
    assert (
        'Error: Invalid parameter value for --parameter '
        f'"{PARAMETER_NO_EQUAL[1]}"' in result.output
    )

    result = isolated_runner.invoke(
        cli, INIT + PARAMETERS_CORRECT + PARAMETER_EARLY_EQUAL
    )
    assert 0 != result.exit_code
    assert (
        'Error: Invalid parameter value for --parameter '
        f'"{PARAMETER_EARLY_EQUAL[1]}"' in result.output
    )

    result = isolated_runner.invoke(cli, INIT + PARAMETERS_CORRECT, INPUT)
    assert 0 == result.exit_code
    assert 'The template requires a value for' in result.output
    for param in set(METADATA.keys()):
        assert param in result.output
    assert (
        'These parameters are not used by the template and were ignored:'
    ) in result.output


@pytest.mark.parametrize('data_dir', ['dir', 'nested/dir/s'])
def test_init_with_data_dir(isolated_runner, data_dir, directory_tree):
    """Test initializing with data directory."""
    from git import Repo

    new_project = Path('test-new-project')
    result = isolated_runner.invoke(cli, INIT + ['--data-dir', data_dir])
    assert 0 == result.exit_code

    assert (new_project / data_dir).exists()
    assert (new_project / data_dir / '.gitkeep').exists()
    assert not Repo(new_project).is_dirty()

    os.chdir(new_project.resolve())
    result = isolated_runner.invoke(
        cli, ['dataset', 'add', '-c', 'my-data', directory_tree.strpath]
    )
    assert 0 == result.exit_code, result.output
    assert (Path(data_dir) / 'my-data' /
            directory_tree.join('file').strpath).exists()


@pytest.mark.parametrize(
    'data_dir', ['/absolute/path/outside', '../relative/path/outside']
)
def test_init_with_wrong_data_dir(isolated_runner, data_dir):
    """Test initialization fails with wrong data directory."""
    result = isolated_runner.invoke(cli, INIT + ['--data-dir', data_dir])
    assert 2 == result.exit_code
    assert f'Data directory {data_dir} is not within project' in result.output


@pytest.mark.parametrize('data_dir', ['.', '.git', '.renku', '.git/'])
def test_init_with_invalid_data_dir(isolated_runner, data_dir):
    """Test initialization fails with invalid data directory."""
    result = isolated_runner.invoke(cli, INIT + ['--data-dir', data_dir])
    assert 2 == result.exit_code
    data_dir = data_dir.rstrip('/')
    assert f'Cannot use {data_dir} as data directory.' in result.output



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test Python SDK client."""

import pytest


def test_local_client(tmpdir):
    """Test a local client."""
    from renku.core.management.client import LocalClient
    client = LocalClient(str(tmpdir.mkdir('project')))

    assert client.path
    assert client.repo is None


@pytest.mark.parametrize(
    'paths, ignored', (
        (['.renku.lock'], ['.renku.lock']),
        (['not ignored', 'lib/foo', 'build/html'], ['lib/foo', 'build/html']),
        (['not ignored'], None),
    )
)
def test_ignored_paths(paths, ignored, client):
    """Test resolution of ignored paths."""
    assert client.find_ignored_paths(*paths) == ignored



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``save`` command."""

from git import Repo

from renku.cli import cli


def test_save_without_remote(runner, project, client, tmpdir_factory):
    """Test saving local changes."""
    with (client.path / 'tracked').open('w') as fp:
        fp.write('tracked file')

    result = runner.invoke(
        cli, ['save', '-m', 'save changes', 'tracked'], catch_exceptions=False
    )
    assert 1 == result.exit_code
    assert 'No remote has been set up' in result.output

    path = str(tmpdir_factory.mktemp('remote'))
    Repo().init(path, bare=True)

    result = runner.invoke(
        cli, ['save', '-d', path, 'tracked'], catch_exceptions=False
    )

    assert 0 == result.exit_code
    assert 'tracked' in result.output
    assert 'Saved changes to: tracked' in client.repo.head.commit.message

    client.repo.delete_remote('origin')


def test_save_with_remote(runner, project, client_with_remote, tmpdir_factory):
    """Test saving local changes."""
    client = client_with_remote['client']
    with (client.path / 'tracked').open('w') as fp:
        fp.write('tracked file')

    result = runner.invoke(
        cli, ['save', '-m', 'save changes', 'tracked'], catch_exceptions=False
    )

    assert 0 == result.exit_code
    assert 'tracked' in result.output
    assert 'save changes' in client.repo.head.commit.message



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent an annotation for a workflow."""

from renku.core.models import jsonld


@jsonld.s(
    type='oa:Annotation',
    context={
        'oa': 'http://www.w3.org/ns/oa#',
        'rdfs': 'http://www.w3.org/2000/01/rdf-schema#',
        'dcterms': 'http://purl.org/dc/terms/',
    },
    cmp=False,
)
class Annotation:
    """Represents a custom annotation for a research object."""

    _id = jsonld.ib(context='@id', kw_only=True)

    body = jsonld.ib(default=None, context='oa:hasBody', kw_only=True)

    source = jsonld.ib(default=None, context='dcterms:creator', kw_only=True)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``log`` command."""
import os

import git
import pytest

from renku.cli import cli


@pytest.mark.shelled
@pytest.mark.parametrize('format', ['json-ld', 'nt', 'rdf'])
def test_run_log_strict(runner, project, run_shell, format):
    """Test log output of run command."""
    # Run a shell command with pipe.
    result = run_shell('renku run echo "a" > output')

    # Assert created output file.
    result = runner.invoke(
        cli, ['log', '--strict', '--format={}'.format(format)]
    )
    assert 0 == result.exit_code, result.output
    assert '.renku/workflow/' in result.output


@pytest.mark.shelled
@pytest.mark.parametrize('format', ['json-ld', 'nt', 'rdf'])
def test_dataset_log_strict(
    tmpdir, runner, project, client, format, subdirectory
):
    """Test output of log for dataset add."""
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code

    paths = []
    test_paths = []
    for i in range(3):
        new_file = tmpdir.join('file_{0}'.format(i))
        new_file.write(str(i))
        paths.append(str(new_file))
        test_paths.append(str(new_file.relto(tmpdir.join('..'))))

    # add data
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'my-dataset'] + paths,
    )
    assert 0 == result.exit_code

    result = runner.invoke(
        cli, ['log', '--strict', '--format={}'.format(format)]
    )

    assert 0 == result.exit_code, result.output
    assert all(p in result.output for p in test_paths)


@pytest.mark.shelled
@pytest.mark.parametrize('format', ['json-ld', 'nt', 'rdf'])
def test_dataset_log_invalidation_strict(
    tmpdir, runner, project, client, format, subdirectory
):
    """Test output of log for dataset add."""
    repo = git.Repo(project)
    input_ = client.path / 'input.txt'
    input_.write_text('first')

    repo.git.add('--all')
    repo.index.commit('Created input.txt')

    os.remove(input_)
    repo.git.add('--all')
    repo.index.commit('Removed input.txt')

    result = runner.invoke(
        cli, ['log', '--strict', '--format={}'.format(format)]
    )

    assert 0 == result.exit_code, result.output
    assert 'wasInvalidatedBy' in result.output



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compatibility layer for different Python versions."""

import cgi
import contextlib
import html
from pathlib import Path

import pathspec

try:
    contextlib.nullcontext
except AttributeError:

    class nullcontext(object):
        """Context manager that does no additional processing.

        Used as a stand-in for a normal context manager, when a particular
        block of code is only sometimes used with a normal context manager:
        cm = optional_cm if condition else nullcontext()
        with cm:
            # Perform operation, using optional_cm if condition is True
        """

        def __init__(self, enter_result=None):
            self.enter_result = enter_result

        def __enter__(self):
            return self.enter_result

        def __exit__(self, *excinfo):
            pass

    contextlib.nullcontext = nullcontext

try:
    FileNotFoundError
except NameError:  # pragma: no cover
    FileNotFoundError = IOError

cgi.escape = html.escape


class RenkuGitWildMatchPattern(pathspec.patterns.GitWildMatchPattern):
    """Custom GitWildMatchPattern matcher."""

    __slots__ = ('pattern', )

    def __init__(self, pattern, include=None):
        """Initialize RenkuRegexPattern."""
        super().__init__(pattern, include)
        self.pattern = pattern


pathspec.util.register_pattern('renku_gitwildmatch', RenkuGitWildMatchPattern)

__all__ = (
    'FileNotFoundError',
    'Path',
    'contextlib',
    'cgi',
)



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service datasets view."""
import os
from pathlib import Path

from flask import Blueprint, request
from flask_apispec import marshal_with, use_kwargs
from git import GitCommandError, Repo

from renku.core.commands.dataset import add_file, create_dataset, \
    dataset_remove, edit_dataset, file_unlink, list_datasets, list_files
from renku.core.commands.save import repo_sync
from renku.core.models import json
from renku.core.utils.contexts import chdir
from renku.service.config import INTERNAL_FAILURE_ERROR_CODE, \
    INVALID_PARAMS_ERROR_CODE, SERVICE_PREFIX
from renku.service.jobs.contexts import enqueue_retry
from renku.service.jobs.datasets import dataset_add_remote_file, dataset_import
from renku.service.jobs.queues import DATASETS_JOB_QUEUE
from renku.service.serializers.datasets import DatasetAddRequest, \
    DatasetAddResponseRPC, DatasetCreateRequest, DatasetCreateResponseRPC, \
    DatasetEditRequest, DatasetEditResponseRPC, DatasetFilesListRequest, \
    DatasetFilesListResponseRPC, DatasetImportRequest, \
    DatasetImportResponseRPC, DatasetListRequest, DatasetListResponseRPC, \
    DatasetRemoveRequest, DatasetRemoveResponseRPC, DatasetUnlinkRequest, \
    DatasetUnlinkResponseRPC
from renku.service.views import error_response, result_response
from renku.service.views.decorators import accepts_json, \
    handle_common_except, header_doc, requires_cache, requires_identity

DATASET_BLUEPRINT_TAG = 'datasets'
dataset_blueprint = Blueprint(
    DATASET_BLUEPRINT_TAG, __name__, url_prefix=SERVICE_PREFIX
)


@use_kwargs(DatasetListRequest, locations=['query'])
@marshal_with(DatasetListResponseRPC)
@header_doc('List all datasets in project.', tags=(DATASET_BLUEPRINT_TAG, ))
@dataset_blueprint.route(
    '/datasets.list',
    methods=['GET'],
    provide_automatic_options=False,
)
@handle_common_except
@requires_cache
@requires_identity
def list_datasets_view(user, cache):
    """List all datasets in project."""
    ctx = DatasetListRequest().load(request.args)
    project = cache.get_project(cache.ensure_user(user), ctx['project_id'])

    with chdir(project.abs_path):
        ctx['datasets'] = list_datasets()

    return result_response(DatasetListResponseRPC(), ctx)


@use_kwargs(DatasetFilesListRequest, locations=['query'])
@marshal_with(DatasetFilesListResponseRPC)
@header_doc('List files in a dataset.', tags=(DATASET_BLUEPRINT_TAG, ))
@dataset_blueprint.route(
    '/datasets.files_list',
    methods=['GET'],
    provide_automatic_options=False,
)
@handle_common_except
@requires_cache
@requires_identity
def list_dataset_files_view(user, cache):
    """List files in a dataset."""
    ctx = DatasetFilesListRequest().load(request.args)
    project = cache.get_project(cache.ensure_user(user), ctx['project_id'])

    with chdir(project.abs_path):
        ctx['files'] = list_files(datasets=[ctx['name']])

    return result_response(DatasetFilesListResponseRPC(), ctx)


@use_kwargs(DatasetAddRequest)
@marshal_with(DatasetAddResponseRPC)
@header_doc(
    'Add an uploaded file to a cloned repository.',
    tags=(DATASET_BLUEPRINT_TAG, )
)
@dataset_blueprint.route(
    '/datasets.add',
    methods=['POST'],
    provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def add_file_to_dataset_view(user_data, cache):
    """Add the uploaded file to cloned repository."""
    ctx = DatasetAddRequest().load(request.json)
    user = cache.ensure_user(user_data)
    project = cache.get_project(user, ctx['project_id'])

    if not ctx['commit_message']:
        ctx['commit_message'] = 'service: dataset add {0}'.format(ctx['name'])

    local_paths = []
    for _file in ctx['files']:
        local_path = None

        if 'file_url' in _file:
            commit_message = '{0}{1}'.format(
                ctx['commit_message'], _file['file_url']
            )

            job = cache.make_job(user)
            _file['job_id'] = job.job_id

            with enqueue_retry(DATASETS_JOB_QUEUE) as queue:
                queue.enqueue(
                    dataset_add_remote_file, user_data, job.job_id,
                    project.project_id, ctx['create_dataset'], commit_message,
                    ctx['name'], _file['file_url']
                )
            continue

        if 'file_id' in _file:
            file = cache.get_file(user, _file['file_id'])
            local_path = file.abs_path

        elif 'file_path' in _file:
            local_path = project.abs_path / Path(_file['file_path'])

        if not local_path or not local_path.exists():
            return error_response(
                INVALID_PARAMS_ERROR_CODE,
                'invalid file reference: {0}'.format(json.dumps(_file))
            )

        ctx['commit_message'] += ' {0}'.format(local_path.name)
        local_paths.append(str(local_path))

    if local_paths:
        with chdir(project.abs_path):
            add_file(
                local_paths,
                ctx['name'],
                create=ctx['create_dataset'],
                force=ctx['force'],
                commit_message=ctx['commit_message']
            )

            try:
                _, ctx['remote_branch'] = repo_sync(
                    Repo(project.abs_path), remote='origin'
                )
            except GitCommandError:
                return error_response(
                    INTERNAL_FAILURE_ERROR_CODE, 'repo sync failed'
                )

    return result_response(DatasetAddResponseRPC(), ctx)


@use_kwargs(DatasetCreateRequest)
@marshal_with(DatasetCreateResponseRPC)
@header_doc(
    'Create a new dataset in a project.', tags=(DATASET_BLUEPRINT_TAG, )
)
@dataset_blueprint.route(
    '/datasets.create',
    methods=['POST'],
    provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def create_dataset_view(user, cache):
    """Create a new dataset in a project."""
    ctx = DatasetCreateRequest().load(request.json)
    project = cache.get_project(cache.ensure_user(user), ctx['project_id'])

    with chdir(project.abs_path):
        create_dataset(
            ctx['name'],
            title=ctx.get('title'),
            creators=ctx.get('creators'),
            description=ctx.get('description'),
            keywords=ctx.get('keywords'),
            commit_message=ctx['commit_message']
        )

    try:
        _, ctx['remote_branch'] = repo_sync(
            Repo(project.abs_path), remote='origin'
        )
    except GitCommandError:
        return error_response(INTERNAL_FAILURE_ERROR_CODE, 'repo sync failed')

    return result_response(DatasetCreateResponseRPC(), ctx)


@use_kwargs(DatasetRemoveRequest)
@marshal_with(DatasetRemoveResponseRPC)
@header_doc('Remove a dataset from a project.', tags=(DATASET_BLUEPRINT_TAG, ))
@dataset_blueprint.route(
    '/datasets.remove',
    methods=['POST'],
    provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def remove_dataset_view(user, cache):
    """Remove a dataset from a project."""
    ctx = DatasetRemoveRequest().load(request.json)
    project = cache.get_project(cache.ensure_user(user), ctx['project_id'])

    if not project.abs_path.exists():
        return error_response(
            INVALID_PARAMS_ERROR_CODE, 'invalid project_id argument'
        )

    with chdir(project.abs_path):
        dataset_remove([ctx['name']], commit_message=ctx['commit_message'])

    try:
        _, ctx['remote_branch'] = repo_sync(
            Repo(project.abs_path), remote='origin'
        )
    except GitCommandError:
        return error_response(INTERNAL_FAILURE_ERROR_CODE, 'repo sync failed')

    return result_response(DatasetRemoveResponseRPC(), ctx)


@use_kwargs(DatasetImportRequest)
@marshal_with(DatasetImportResponseRPC)
@header_doc('Import a dataset', tags=(DATASET_BLUEPRINT_TAG, ))
@dataset_blueprint.route(
    '/datasets.import',
    methods=['POST'],
    provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def import_dataset_view(user_data, cache):
    """Import a dataset view."""
    user = cache.ensure_user(user_data)
    ctx = DatasetImportRequest().load(request.json)
    project = cache.get_project(user, ctx['project_id'])
    job = cache.make_job(user, locked=project.project_id)

    with enqueue_retry(DATASETS_JOB_QUEUE) as queue:
        queue.enqueue(
            dataset_import,
            user_data,
            job.job_id,
            project.project_id,
            ctx['dataset_uri'],
            name=ctx.get('name'),
            extract=ctx.get('extract', False),
            timeout=int(os.getenv('WORKER_DATASET_JOBS_TIMEOUT', 1800)),
            result_ttl=int(os.getenv('WORKER_DATASET_JOBS_RESULT_TTL', 500))
        )

    return result_response(DatasetImportResponseRPC(), job)


@use_kwargs(DatasetEditRequest)
@marshal_with(DatasetEditResponseRPC)
@header_doc('Edit dataset metadata', tags=(DATASET_BLUEPRINT_TAG, ))
@dataset_blueprint.route(
    '/datasets.edit',
    methods=['POST'],
    provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def edit_dataset_view(user_data, cache):
    """Edit dataset metadata."""
    ctx = DatasetEditRequest().load(request.json)

    user = cache.ensure_user(user_data)
    project = cache.get_project(user, ctx['project_id'])

    if ctx.get('commit_message') is None:
        ctx['commit_message'] = 'service: dataset edit {0}'.format(ctx['name'])

    with chdir(project.abs_path):
        edited, warnings = edit_dataset(
            ctx['name'],
            ctx.get('title'),
            ctx.get('description'),
            ctx.get('creators'),
            keywords=ctx.get('keywords'),
            commit_message=ctx['commit_message']
        )

    return result_response(
        DatasetEditResponseRPC(), {
            'edited': {field: ctx.get(field)
                       for field in edited},
            'warnings': warnings
        }
    )


@use_kwargs(DatasetUnlinkRequest)
@marshal_with(DatasetUnlinkResponseRPC)
@header_doc('Unlink a file from a dataset', tags=(DATASET_BLUEPRINT_TAG, ))
@dataset_blueprint.route(
    '/datasets.unlink',
    methods=['POST'],
    provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def unlink_file_view(user_data, cache):
    """Unlink a file from a dataset."""
    ctx = DatasetUnlinkRequest().load(request.json)

    include = ctx.get('include_filter')
    exclude = ctx.get('exclude_filter')

    user = cache.ensure_user(user_data)
    project = cache.get_project(user, ctx['project_id'])

    if ctx.get('commit_message') is None:
        if include and exclude:
            filters = '-I {0} -X {0}'.format(include, exclude)
        elif not include and exclude:
            filters = '-X {0}'.format(exclude)
        else:
            filters = '-I {0}'.format(include)

        ctx['commit_message'] = (
            'service: unlink dataset {0} {1}'.format(ctx['name'], filters)
        )

    with chdir(project.abs_path):
        records = file_unlink(
            name=ctx['name'],
            include=ctx.get('include_filters'),
            exclude=ctx.get('exclude_filters'),
            yes=True,
            interactive=False,
            commit_message=ctx['commit_message']
        )

        unlinked = [record.path for record in records]

    return result_response(DatasetUnlinkResponseRPC(), {'unlinked': unlinked})



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migrate datasets with type scoped contexts for support with pyld 2.0.

Since pyld now doesn't allow ':' in type scoped contexts and this is not
backwards compatible, we need to add/duplicate this migrations as a sort of
back-dated migration.
"""

import re


def migrate(client):
    """Migration function."""
    _migrate_datasets(client)


def _migrate_datasets(client):
    """Migrate type scoped contexts of datasets."""
    paths = (client.path / client.renku_datasets_path).rglob(client.METADATA)
    for path in paths:
        with path.open('r') as dataset:
            content = dataset.read()

        content = re.sub(r'"([^"])+_prov:([^"]+)":', '"\1_prov_\2":', content)
        content = re.sub(
            r'"([^"])+_wfprov:([^"]+)":', '"\1_wfprov_\2":', content
        )
        content = re.sub(
            r'"([^"])+_schema:([^"]+)":', '"\1_schema_\2":', content
        )

        with path.open('w') as dataset:
            dataset.write(content)



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

import pytest

from renku.core.models.cwl.command_line_tool import CommandLineToolFactory
from renku.core.models.entities import Collection, Entity


def test_1st_tool(client):
    """Check creation of 1st tool example from args."""
    tool = CommandLineToolFactory(('echo', 'Hello world!')
                                  ).generate_process_run(
                                      client=client,
                                      commit=client.repo.head.commit,
                                      path='dummy.yaml'
                                  )

    tool = tool.association.plan
    assert 'Hello world!' == tool.arguments[0].value


def test_03_input(client):
    """Check the essential input parameters."""
    whale = Path(client.path) / 'whale.txt'
    whale.touch()

    client.repo.index.add([str(whale)])
    client.repo.index.commit('add whale.txt')

    argv = [
        'echo',
        '-f',
        '-i42',
        '--example-string',
        'hello',
        '--file=whale.txt',
    ]
    tool = CommandLineToolFactory(
        argv, directory=client.path, working_dir=client.path
    ).generate_process_run(
        client=client, commit=client.repo.head.commit, path='dummy.yaml'
    )

    tool = tool.association.plan

    assert ['-f'] == tool.arguments[0].to_argv()

    assert 42 == tool.arguments[1].value
    assert '-i' == tool.arguments[1].prefix

    assert 'hello' == tool.arguments[2].value
    assert '--example-string ' == tool.arguments[2].prefix

    assert tool.inputs[0].consumes.path == 'whale.txt'
    assert isinstance(tool.inputs[0].consumes, Entity)
    assert not isinstance(tool.inputs[0].consumes, Collection)
    assert '--file=' == tool.inputs[0].prefix

    assert argv == tool.to_argv()


def test_base_command_detection(client):
    """Test base command detection."""
    hello = Path(client.path) / 'hello.tar'
    hello.touch()

    client.repo.index.add([str(hello)])
    client.repo.index.commit('add hello.tar')

    argv = ['tar', 'xf', 'hello.tar']
    tool = CommandLineToolFactory(
        argv, directory=client.path, working_dir=client.path
    ).generate_process_run(
        client=client, commit=client.repo.head.commit, path='dummy.yaml'
    )

    tool = tool.association.plan

    assert 'tar xf' == tool.command
    assert tool.inputs[0].consumes.path == 'hello.tar'
    assert isinstance(tool.inputs[0].consumes, Entity)
    assert not isinstance(tool.inputs[0].consumes, Collection)
    assert tool.inputs[0].prefix is None

    assert argv == tool.to_argv()


def test_base_command_as_file_input(client):
    """Test base command detection when it is a script file."""
    cwd = Path(client.path)
    script = cwd / 'script.py'
    script.touch()

    input_file = cwd / 'input.csv'
    input_file.touch()

    client.repo.index.add([str(script), str(input_file)])
    client.repo.index.commit('add file')

    argv = ['script.py', 'input.csv']
    tool = CommandLineToolFactory(
        argv,
        directory=client.path,
        working_dir=client.path,
    ).generate_process_run(
        client=client, commit=client.repo.head.commit, path='dummy.yaml'
    )

    tool = tool.association.plan

    assert not tool.command
    assert 2 == len(tool.inputs)


def test_short_base_command_detection(client):
    """Test base command detection without arguments."""
    tool = CommandLineToolFactory(('echo', 'A')).generate_process_run(
        client=client, commit=client.repo.head.commit, path='dummy.yaml'
    )

    tool = tool.association.plan

    assert 'A' == tool.arguments[0].value
    assert ['echo', 'A'] == tool.to_argv()


def test_04_output(client):
    """Test describtion of outputs from a command."""
    hello = Path(client.path) / 'hello.tar'
    hello.touch()

    client.repo.index.add([str(hello)])
    client.repo.index.commit('add hello.tar')

    argv = ['tar', 'xf', 'hello.tar']
    factory = CommandLineToolFactory(
        argv, directory=client.path, working_dir=client.path
    )

    # simulate run

    output = Path(client.path) / 'hello.txt'
    output.touch()

    parameters = list(factory.guess_outputs([output]))

    assert 'File' == parameters[0][0].type
    assert 'hello.txt' == parameters[0][0].outputBinding.glob

    tool = factory.generate_process_run(
        client=client, commit=client.repo.head.commit, path='dummy.yaml'
    )

    tool = tool.association.plan
    assert argv == tool.to_argv()


def test_05_stdout(client):
    """Test stdout mapping."""
    output = Path(client.path) / 'output.txt'
    output.touch()

    client.repo.index.add([str(output)])
    client.repo.index.commit('add output')

    argv = ['echo', 'Hello world!']
    factory = CommandLineToolFactory(
        argv,
        directory=client.path,
        working_dir=client.path,
        stdout='output.txt',
    )

    assert 'output.txt' == factory.stdout
    assert 'stdout' == factory.outputs[0].type

    tool = factory.generate_process_run(
        client=client, commit=client.repo.head.commit, path='dummy.yaml'
    )

    tool = tool.association.plan
    assert argv == tool.to_argv()


def test_stdout_with_conflicting_arg(client):
    """Test stdout with conflicting argument value."""
    output = Path(client.path) / 'lalala'
    output.touch()

    client.repo.index.add([str(output)])
    client.repo.index.commit('add lalala')

    argv = ['echo', 'lalala']
    factory = CommandLineToolFactory(
        argv,
        directory=client.path,
        working_dir=client.path,
        stdout='lalala',
    )

    assert 'lalala' == factory.inputs[0].default
    assert 'string' == factory.inputs[0].type
    assert 'lalala' == factory.stdout
    assert 'stdout' == factory.outputs[0].type

    tool = factory.generate_process_run(
        client=client, commit=client.repo.head.commit, path='dummy.yaml'
    )

    tool = tool.association.plan
    assert argv == tool.to_argv()


def test_06_params(client):
    """Test referencing input parameters in other fields."""
    hello = Path(client.path) / 'hello.tar'
    hello.touch()
    client.repo.index.add([str(hello)])
    client.repo.index.commit('add hello.tar')

    argv = ['tar', 'xf', 'hello.tar', 'goodbye.txt']
    factory = CommandLineToolFactory(
        argv,
        directory=client.path,
        working_dir=client.path,
    )

    assert 'goodbye.txt' == factory.inputs[1].default
    assert 'string' == factory.inputs[1].type
    assert 2 == factory.inputs[1].inputBinding.position

    goodbye_id = factory.inputs[1].id

    # simulate run

    output = Path(client.path) / 'goodbye.txt'
    output.touch()

    parameters = list(factory.guess_outputs([output]))

    assert 'File' == parameters[0][0].type
    assert '$(inputs.{0})'.format(goodbye_id) == \
           parameters[0][0].outputBinding.glob

    tool = factory.generate_process_run(
        client=client, commit=client.repo.head.commit, path='dummy.yaml'
    )

    tool = tool.association.plan
    assert argv == tool.to_argv()


def test_09_array_inputs(client):
    """Test specification of input parameters in arrays."""
    argv = [
        'echo',
        '-A',
        'one',
        'two',
        'three',
        '-B=four',
        '-B=five',
        '-B=six',
        '-C=seven,eight,nine',
    ]
    tool = CommandLineToolFactory(
        argv, directory=client.path, working_dir=client.path
    ).generate_process_run(
        client=client, commit=client.repo.head.commit, path='dummy.yaml'
    )

    tool = tool.association.plan

    assert 'seven,eight,nine' == tool.arguments[-1].value
    assert '-C=' == tool.arguments[-1].prefix

    assert argv == tool.to_argv()


@pytest.mark.parametrize('argv', [['wc'], ['wc', '-l']])
def test_stdin_and_stdout(argv, client):
    """Test stdout mapping."""
    input_ = Path(client.path) / 'input.txt'
    input_.touch()
    output = Path(client.path) / 'output.txt'
    output.touch()
    error = Path(client.path) / 'error.log'
    error.touch()

    client.repo.index.add([str(input_), str(output), str(error)])
    client.repo.index.commit('add files')

    factory = CommandLineToolFactory(
        argv,
        directory=client.path,
        working_dir=client.path,
        stdin='input.txt',
        stdout='output.txt',
        stderr='error.log',
    )

    assert factory.stdin
    if len(argv) > 1:
        assert factory.arguments

    assert 'output.txt' == factory.stdout
    assert 'stdout' == factory.outputs[0].type

    tool = factory.generate_process_run(
        client=client, commit=client.repo.head.commit, path='dummy.yaml'
    )

    tool = tool.association.plan
    assert argv == tool.to_argv()
    assert any(
        i.mapped_to and i.mapped_to.stream_type == 'stdin' for i in tool.inputs
    )
    assert any(
        o.mapped_to and o.mapped_to.stream_type == 'stdout'
        for o in tool.outputs
    )
    assert any(
        o.mapped_to and o.mapped_to.stream_type == 'stderr'
        for o in tool.outputs
    )


def test_input_directory(client):
    """Test input directory."""
    cwd = Path(client.path)
    src = cwd / 'src'
    src.mkdir(parents=True)

    for i in range(5):
        (src / str(i)).touch()

    src_tar = cwd / 'src.tar'
    src_tar.touch()

    client.repo.index.add([str(src), str(src_tar)])
    client.repo.index.commit('add file and folder')

    argv = ['tar', 'czvf', 'src.tar', 'src']
    factory = CommandLineToolFactory(
        argv,
        directory=client.path,
        working_dir=client.path,
    )

    tool = factory.generate_process_run(
        client=client, commit=client.repo.head.commit, path='dummy.yaml'
    )

    tool = tool.association.plan
    assert argv == tool.to_argv()

    inputs = sorted(tool.inputs, key=lambda x: x.position)

    assert src_tar.name == inputs[0].consumes.path
    assert isinstance(inputs[0].consumes, Entity)
    assert not isinstance(inputs[0].consumes, Collection)
    assert inputs[1].consumes.path == src.name
    assert isinstance(inputs[1].consumes, Collection)


def test_existing_output_directory(client, runner, project):
    """Test creation of InitialWorkDirRequirement for output."""
    from renku.core.models.workflow.converters.cwl import CWLConverter

    client.path = client.path
    output = client.path / 'output'

    argv = ['script', 'output']
    factory = CommandLineToolFactory(
        argv,
        directory=client.path,
        working_dir=client.path,
    )

    with factory.watch(client, no_output=True) as tool:
        # Script creates the directory.
        output.mkdir(parents=True)

    run = factory.generate_process_run(
        client=client, commit=client.repo.head.commit, path='dummy.yaml'
    )

    cwl, _ = CWLConverter.convert(run.association.plan, client)

    assert 0 == len([r for r in cwl.requirements if hasattr(r, 'listing')])

    output.mkdir(parents=True, exist_ok=True)
    with factory.watch(client) as tool:
        # The directory already exists.
        (output / 'result.txt').touch()

    assert 1 == len(tool.inputs)

    run = tool.generate_process_run(
        client=client, commit=client.repo.head.commit, path='dummy.yaml'
    )
    cwl, _ = CWLConverter.convert(run.association.plan, client)

    reqs = [r for r in cwl.requirements if hasattr(r, 'listing')]

    assert 1 == len(reqs)
    assert output.name == reqs[0].listing[0].entryname
    assert 1 == len(tool.outputs)



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Initial migrations."""
import os
import shutil
import urllib
import uuid
from pathlib import Path, posixpath
from urllib.parse import quote

from renku.core.management.repository import DEFAULT_DATA_DIR as DATA_DIR
from renku.core.models.datasets import Dataset
from renku.core.models.refs import LinkReference
from renku.core.utils.urls import url_to_string


def migrate(client):
    """Migration function."""
    _ensure_clean_lock(client)
    _do_not_track_lock_file(client)
    _migrate_datasets_pre_v0_3(client)
    _migrate_broken_dataset_paths(client)
    _fix_uncommitted_labels(client)
    _fix_dataset_files_urls(client)
    _fix_broken_dataset_file_project(client)
    _dataset_file_id_migration(client)
    _migrate_files_project(client)


def _ensure_clean_lock(client):
    """Make sure Renku lock file is not part of repository."""
    lock_file = client.path / '.renku.lock'
    try:
        lock_file.unlink()
    except FileNotFoundError:
        pass


def _do_not_track_lock_file(client):
    """Add lock file to .gitingore if not already exists."""
    # Add lock file to .gitignore.
    lock_file = '.renku.lock'
    gitignore = client.path / '.gitignore'
    if lock_file not in gitignore.read_text():
        gitignore.open('a').write('\n{0}\n'.format(lock_file))


def _migrate_datasets_pre_v0_3(client):
    """Migrate datasets from Renku 0.3.x."""

    def _dataset_pre_0_3(client):
        """Return paths of dataset metadata for pre 0.3.4."""
        project_is_pre_0_3 = int(client.project.version) < 2
        if project_is_pre_0_3:
            return (client.path / DATA_DIR).rglob(client.METADATA)
        return []

    for old_path in _dataset_pre_0_3(client):
        name = str(old_path.parent.relative_to(client.path / DATA_DIR))

        dataset = Dataset.from_yaml(old_path, client=client)
        new_path = (client.renku_datasets_path / dataset.uid / client.METADATA)
        new_path.parent.mkdir(parents=True, exist_ok=True)

        with client.with_metadata(read_only=True) as meta:
            for module in client.repo.submodules:
                if Path(module.url).name == meta.name:
                    module.remove()

        for file_ in dataset.files:
            if not Path(file_.path).exists():
                expected_path = (
                    client.path / DATA_DIR / dataset.name / file_.path
                )
                if expected_path.exists():
                    file_.path = expected_path.relative_to(client.path)

        dataset.__reference__ = new_path.relative_to(client.path)
        dataset.to_yaml()

        Path(old_path).unlink()
        ref = LinkReference.create(
            client=client,
            name='datasets/{0}'.format(name),
            force=True,
        )
        ref.set_reference(new_path)


def _migrate_broken_dataset_paths(client):
    """Ensure all paths are using correct directory structure."""
    for dataset in client.datasets.values():
        dataset_path = client.path / dataset.path

        expected_path = (
            client.renku_datasets_path /
            Path(quote(dataset.identifier, safe=''))
        )

        # migrate the refs
        ref = LinkReference.create(
            client=client,
            name='datasets/{0}'.format(dataset.name),
            force=True,
        )
        ref.set_reference(expected_path / client.METADATA)

        if not dataset_path.exists():
            dataset_path = (
                client.renku_datasets_path / uuid.UUID(dataset.identifier).hex
            )

        if not expected_path.exists():
            shutil.move(dataset_path, expected_path)
            dataset.path = expected_path
            dataset.__reference__ = expected_path / client.METADATA

        for file_ in dataset.files:
            file_path = Path(file_.path)
            if not file_path.exists() and file_.path.startswith('..'):
                new_path = (
                    client.renku_datasets_path / dataset.uid / file_path
                ).resolve().relative_to(client.path)

                file_.path = new_path

                _, commit, _ = client.resolve_in_submodules(
                    client.find_previous_commit(file_.path, revision='HEAD'),
                    file_.path,
                )
                host = client.remote.get('host') or 'localhost'
                host = os.environ.get('RENKU_DOMAIN') or host

                # always set the id by the identifier
                file_._id = urllib.parse.urljoin(
                    'https://{host}'.format(host=host),
                    posixpath.join(
                        '/blob/{hexsha}/{path}'.format(
                            hexsha=commit.hexsha, path=new_path
                        )
                    )
                )
                file_._label = '{}@{}'.format(new_path, commit.hexsha)

        dataset.to_yaml()


def _fix_uncommitted_labels(client):
    """Ensure files have correct label instantiation."""
    for dataset in client.datasets.values():
        for file_ in dataset.files:
            try:
                _, commit, _ = client.resolve_in_submodules(
                    client.find_previous_commit(file_.path, revision='HEAD'),
                    file_.path,
                )
                file_.commit = commit
                if (
                    not file_._label or 'UNCOMMITTED' in file_._label or
                    '@' not in file_._label
                ):
                    file_._label = file_.default_label()
                    file_._id = file_.default_id()
            except KeyError:
                pass
        dataset.to_yaml()


def _fix_dataset_files_urls(client):
    """Ensure dataset files have correct url format."""
    for dataset in client.datasets.values():
        for file_ in dataset.files:
            if file_.url:
                file_.url = url_to_string(file_.url)

        dataset.to_yaml()


def _fix_broken_dataset_file_project(client):
    """Ensure project is correctly set on ``DatasetFile``."""
    for dataset in client.datasets.values():
        for file_ in dataset.files:
            if not file_._project or 'NULL/NULL' in file_._project._id:
                file_._project = client.project

        dataset.to_yaml()


def _dataset_file_id_migration(client):
    """Ensure dataset files have a fully qualified url."""
    for dataset in client.datasets.values():
        for file_ in dataset.files:
            if not file_._id.startswith('https'):
                file_._id = file_.default_id()

        dataset.to_yaml()


def _migrate_files_project(client):
    """Ensure dataset files have correct project."""
    for dataset in client.datasets.values():
        for file_ in dataset.files:
            file_._project = dataset._project

        dataset.to_yaml()
